# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import multiprocessing as mp
import os
import time

import numpy as np
import pandas as pd
import tensorflow as tf

from fastestimator.util.op import get_op_from_mode, verify_ops, write_outputs_by_key, get_inputs_by_op


class RecordWriter:
    def __init__(self,
                 train_data,
                 save_dir,
                 validation_data=None,
                 ops=None,
                 write_feature=None,
                 expand_dims=False,
                 max_record_size_mb=300,
                 compression=None):
        self.train_data = train_data
        self.save_dir = save_dir
        self.validation_data = validation_data
        self.ops = ops
        self.write_feature = write_feature
        self.expand_dims = expand_dims
        self.max_record_size_mb = max_record_size_mb
        self.compression = compression
        self.num_process = mp.cpu_count()
        self.compression_option = tf.io.TFRecordOptions(compression_type=compression)
        self.global_file_idx = {"train": 0, "eval": 0}
        self.global_feature_key = {"train": [], "eval": []}
        self.feature_set_idx = 0
        self._verify_inputs()

    def _verify_inputs(self):
        if any(isinstance(inp, tuple) for inp in [self.train_data, self.validation_data, self.ops, self.write_feature]):
            num_unpaired_feature_sets = [len(self.train_data)]
            if self.validation_data:
                assert isinstance(self.validation_data,
                                  tuple), "validation data must be tuple when creating unpaired feature set"
                num_unpaired_feature_sets.append(len(self.validation_data))
            else:
                self.validation_data = [None] * len(self.train_data)
            if self.ops:
                assert isinstance(self.ops, tuple), "operation must be tuple when creating unpaired feature set"
                num_unpaired_feature_sets.append(len(self.ops))
            else:
                self.ops = [None] * len(self.train_data)
            if self.write_feature:
                assert isinstance(self.write_feature,
                                  tuple), "write_feature must be tuple when creating unpaired feature set"
                num_unpaired_feature_sets.append((len(self.write_feature)))
            else:
                self.write_feature = [None] * len(self.train_data)
            assert len(set(
                num_unpaired_feature_sets)) == 1, "tuple length should be consistent when creating unpaired feature set"
            self.train_data, self.validation_data, self.ops, self.write_feature = list(self.train_data), list(
                self.validation_data), list(self.ops), list(self.write_feature)
        else:
            self.train_data, self.validation_data, self.ops, self.write_feature = \
                [self.train_data], [self.validation_data], [self.ops], [self.write_feature]
        for idx in range(len(self.train_data)):
            assert type(self.train_data[idx]) is dict or self.train_data[idx].endswith(
                ".csv"), "train data should either be a dictionary or a csv path"
            if self.validation_data[idx]:
                assert type(self.validation_data[idx]) in [dict, float] or self.validation_data[idx].endswith(
                    ".csv"), "validation data supports partition ratio (float), csv file or dictionary"
            if self.write_feature[idx]:
                assert isinstance(self.write_feature[idx],
                                  (list, dict)), "write_feature must be either list or dictionary"
            if self.ops[idx]:
                if not isinstance(self.ops, list):
                    self.ops[idx] = [self.ops[idx]]
            else:
                self.ops[idx] = []

    @staticmethod
    def _int64_feature(value):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

    @staticmethod
    def _bytes_feature(value):
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

    @staticmethod
    def _float_feature(value):
        return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))

    def write(self):
        self._prepare_savepath(self.save_dir)
        for train_data, validation_data, write_feature, ops in zip(self.train_data, self.validation_data,
                                                                   self.write_feature, self.ops):
            self._create_record_local(train_data, validation_data, write_feature, ops)
            self.feature_set_idx += 1

    def _create_record_local(self, train_data, validation_data, write_feature, ops):
        self.num_example_csv, self.num_example_record, self.mb_per_csv_example, self.mb_per_record_example = \
            {}, {}, {}, {}
        self.mode_ops, self.feature_name, self.feature_dtype, self.feature_shape = {}, {}, {}, {}
        self.train_data_local, self.validation_data_local, self.write_feature_local, self.ops_local = \
            train_data, validation_data, write_feature, ops
        self._prepare_train()
        self._get_feature_info(self.train_data_local, "train")
        if validation_data:
            self._prepare_validation()
            self._get_feature_info(self.validation_data_local, "eval")
        self.num_example_record["train"] = self._write_tfrecord_parallel(self.train_data_local, mode="train")
        self._write_json_summary("train")
        if validation_data:
            self.num_example_record["eval"] = self._write_tfrecord_parallel(self.validation_data_local, mode="eval")
            self._write_json_summary("eval")

    def _get_feature_info(self, dictionary, mode):
        feature = self._transform_one_slice(dictionary=dictionary, index=0, mode=mode)
        if self.write_feature_local is None:
            self.feature_name[mode] = list(feature.keys())
        elif isinstance(self.write_feature_local, dict):
            self.feature_name[mode] = self.write_feature_local[mode]
        else:
            self.feature_name[mode] = self.write_feature_local
        self.global_feature_key[mode].extend(self.feature_name[mode])
        assert len(set(self.global_feature_key[mode])) == len(
            self.global_feature_key[mode]), "found duplicate key in feature name during {}: {}".format(
            mode, self.global_feature_key[mode])
        self.mb_per_csv_example[mode] = 0
        self.mb_per_record_example[mode] = 0
        self.feature_dtype[mode] = {}
        self.feature_shape[mode] = {}
        for key in self.feature_name[mode]:
            data = np.asarray(feature[key])
            self.mb_per_csv_example[mode] += data.nbytes / 1e6
            if self.expand_dims:
                data = data[0]
            self.mb_per_record_example[mode] += data.nbytes / 1e6
            dtype = str(data.dtype)
            if "<U" in dtype:
                dtype = "str"
            self.feature_dtype[mode][key] = dtype
            if data.size == 1:
                self.feature_shape[mode][key] = []
            elif max(data.shape) == np.prod(data.shape):
                self.feature_shape[mode][key] = [-1]
            else:
                self.feature_shape[mode][key] = data.shape

    def _write_tfrecord_parallel(self, dictionary, mode):
        num_example_list = []
        feature_shape_list = []
        processes = []
        queue_example = mp.Queue()
        queue_shape = mp.Queue()
        num_files_process = int(
            np.ceil(self.num_example_csv[mode] * self.mb_per_csv_example[mode] / self.max_record_size_mb /
                    self.num_process))
        num_example_process_remain = self.num_example_csv[mode] % self.num_process
        serial_start = 0
        file_idx_start = 0
        for i in range(self.num_process):
            if i < num_example_process_remain:
                num_example_process = self.num_example_csv[mode] // self.num_process + 1
            else:
                num_example_process = self.num_example_csv[mode] // self.num_process
            serial_end = serial_start + num_example_process
            processes.append(
                mp.Process(
                    target=self._write_tfrecord_serial,
                    args=(dictionary,
                          serial_start,
                          serial_end,
                          num_files_process,
                          file_idx_start,
                          mode,
                          queue_example,
                          queue_shape)))
            serial_start += num_example_process
            file_idx_start += num_files_process
        for p in processes:
            p.start()
        for _ in range(self.num_process):
            num_example_list.extend(queue_example.get(block=True))
            feature_shape_list.append(queue_shape.get(block=True))
        for p in processes:
            p.join()
        self._reconfirm_shape(feature_shape_list, mode)
        return num_example_list

    def _reconfirm_shape(self, feature_shape_list, mode):
        feature_shape = self.feature_shape[mode]
        for new_feature_shape in feature_shape_list:
            for key in feature_shape:
                if new_feature_shape[key] == [-1] and feature_shape[key] == []:
                    feature_shape[key] = [-1]
        self.feature_shape[mode] = feature_shape

    def _write_tfrecord_serial(self,
                               dictionary,
                               serial_start,
                               serial_end,
                               num_files_process,
                               file_idx_start,
                               mode,
                               queue_example,
                               queue_shape):
        num_example_list = []
        num_csv_example_per_file = (serial_end - serial_start) // num_files_process
        show_progress = serial_start == 0
        file_start = serial_start
        file_end = file_start + num_csv_example_per_file
        for i in range(num_files_process):
            file_idx = i + file_idx_start + self.global_file_idx[mode]
            filename = mode + str(file_idx) + ".tfrecord"
            if i == (num_files_process - 1):
                file_end = serial_end
            num_example_list.append(
                self._write_single_file(dictionary,
                                        filename,
                                        file_start,
                                        file_end,
                                        serial_start,
                                        serial_end,
                                        show_progress,
                                        mode))
            file_start += num_csv_example_per_file
            file_end += num_csv_example_per_file
        queue_example.put(num_example_list)
        queue_shape.put(self.feature_shape[mode])

    def _write_single_file(self,
                           dictionary,
                           filename,
                           file_start,
                           file_end,
                           serial_start,
                           serial_end,
                           show_progress,
                           mode):
        goal_number = serial_end - serial_start
        logging_interval = max(goal_number // 20, 1)
        num_example = 0
        with tf.io.TFRecordWriter(os.path.join(self.save_dir, filename), options=self.compression_option) as writer:
            for i in range(file_start, file_end):
                if i == file_start and show_progress:
                    time_start = time.perf_counter()
                    example_start = num_example
                if (i - serial_start) % logging_interval == 0 and show_progress:
                    if i == 0:
                        record_per_sec = 0.0
                    else:
                        record_per_sec = (num_example - example_start) * self.num_process / (time.perf_counter() -
                                                                                             time_start)
                    print("FastEstimator: Converting %s TFRecords %.1f%%, Speed: %.2f record/sec" %
                          (mode.capitalize(), (i - serial_start) / goal_number * 100, record_per_sec))
                feature = self._transform_one_slice(dictionary, i, mode=mode)
                if self.expand_dims:
                    num_patches = self._verify_dict(feature, mode)
                    for j in range(num_patches):
                        feature_patch = self._get_dict_slice(feature, j)
                        self._write_single_example(feature_patch, writer, mode)
                        num_example += 1
                else:
                    self._write_single_example(feature, writer, mode)
                    num_example += 1
        return filename, num_example

    def _transform_one_slice(self, dictionary, index, mode):
        feature = self._get_dict_slice(dictionary, index)
        if self.ops_local:
            feature = self._preprocessing(feature, mode)
        return feature

    def _write_single_example(self, dictionary, writer, mode):
        feature_tfrecord = {}
        for key in self.feature_name[mode]:
            data = np.array(dictionary[key]).astype(self.feature_dtype[mode][key])
            expected_shape = self.feature_shape[mode][key]
            assert data.size > 0, "found empty data on feature '{}'".format(key)
            if len(expected_shape) > 1:
                assert expected_shape == data.shape, \
                    "inconsistent shape on same feature `{}` among different examples, expected `{}`, found `{}`" \
                    .format(key, expected_shape, data.shape)
            else:
                if data.size > 1:
                    assert max(data.shape) == np.prod(data.shape), \
                        "inconsistent shape on same feature `{}` among different examples,\
                         expected 0 or 1 dimensional array, found `{}`".format(key, data.shape)
                    if not expected_shape:
                        self.feature_shape[mode][key] = [-1]
            feature_tfrecord[key] = self._bytes_feature(data.tostring())
        example = tf.train.Example(features=tf.train.Features(feature=feature_tfrecord))
        writer.write(example.SerializeToString())

    @staticmethod
    def _get_dict_slice(dictionary, index):
        feature_slice = dict()
        keys = dictionary.keys()
        for key in keys:
            feature_slice[key] = dictionary[key][index]
        return feature_slice

    def _preprocessing(self, feature, mode):
        data = None
        for op in self.mode_ops[mode]:
            data = get_inputs_by_op(op, feature, data)
            data = op.forward(data, state={"mode": mode})
            if op.outputs:
                feature = write_outputs_by_key(feature, data, op.outputs)
        return feature

    def _verify_dict(self, dictionary, mode=None):
        if mode is None:
            feature_name = dictionary.keys()
        else:
            feature_name = self.feature_name[mode]
        num_example_list = []
        for key in feature_name:
            feature_data = dictionary[key]
            if type(feature_data) is list:
                num_example_list.append(len(feature_data))
            elif type(feature_data) is np.ndarray:
                num_example_list.append(feature_data.shape[0])
            else:
                raise ValueError("the feature only supports list or numpy array")
        assert len(set(num_example_list)) == 1, "features should have the same number of examples"
        return set(num_example_list).pop()

    def _prepare_train(self):
        if type(self.train_data_local) is str:
            df = pd.read_csv(self.train_data_local)
            self.train_data_local = df.to_dict('list')
        self.num_example_csv["train"] = self._verify_dict(self.train_data_local)
        self._check_ops("train")

    def _prepare_validation(self):
        if type(self.validation_data_local) is float:
            num_example_takeout = int(self.validation_data_local * self.num_example_csv["train"])
            train_idx = range(self.num_example_csv["train"])
            eval_idx = np.random.choice(train_idx, num_example_takeout, replace=False)
            train_idx = np.delete(train_idx, eval_idx)
            self.validation_data_local = {}
            for key in self.train_data_local.keys():
                total_data = self.train_data_local[key]
                if type(total_data) is list:
                    self.train_data_local[key] = [total_data[x] for x in train_idx]
                    self.validation_data_local[key] = [total_data[x] for x in eval_idx]
                else:
                    self.train_data_local[key] = total_data[train_idx]
                    self.validation_data_local[key] = total_data[eval_idx]
            self.num_example_csv["train"] = self.num_example_csv["train"] - num_example_takeout
        elif type(self.validation_data_local) is str:
            df = pd.read_csv(self.validation_data_local)
            self.validation_data_local = df.to_dict('list')
        self.num_example_csv["eval"] = self._verify_dict(self.validation_data_local)
        self._check_ops("eval")

    def _check_ops(self, mode):
        if self.ops_local:
            self.mode_ops[mode] = get_op_from_mode(self.ops_local, mode)
            if len(self.mode_ops[mode]) > 0:
                verify_ops(self.mode_ops[mode], "RecordWriter")

    def _prepare_savepath(self, save_dir):
        self.save_dir = save_dir
        if os.path.exists(self.save_dir):
            assert len(os.listdir(self.save_dir)) == 0, "Cannot save to {} because the direcotry is not empty".format(
                self.save_dir)
        else:
            os.makedirs(self.save_dir)
        print("FastEstimator: Saving tfrecord to %s" % self.save_dir)

    def _write_json_summary(self, mode):
        summary = {"feature_dtype": self.feature_dtype[mode], "feature_shape": self.feature_shape[mode]}
        files, num_examples = zip(*self.num_example_record[mode])
        summary["file_names"] = list(files)
        summary["num_examples"] = list(num_examples)
        summary["example_size_mb"] = self.mb_per_record_example[mode]
        if self.compression:
            summary["compression"] = self.compression
        file_name = "%s_summary%d.json" % (mode, self.feature_set_idx)
        with open(os.path.join(self.save_dir, file_name), 'w') as fp:
            json.dump(summary, fp, indent=4)
        self.global_file_idx[mode] += len(files)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


class TensorOp:
    def __init__(self, inputs=None, outputs=None, mode=None):
        self.inputs = inputs
        self.outputs = outputs
        self.mode = mode

    def forward(self, data, state):
        return data


class NumpyOp:
    def __init__(self, inputs=None, outputs=None, mode=None):
        self.inputs = inputs
        self.outputs = outputs
        self.mode = mode

    def forward(self, data, state):
        return data


def get_op_from_mode(ops, current_mode):
    selected_ops = []
    for op in ops:
        assert hasattr(op, "mode"), "Operation: {} has no mode attribute".format(op)
        op_mode = op.mode
        if not isinstance(op_mode, list):
            op_mode = [op_mode]
        if None in op_mode or current_mode in op_mode:
            selected_ops.append(op)
    return selected_ops


def get_inputs_by_op(op, store, default=None):
    data = default
    if op.inputs:
        if hasattr(op.inputs, "__call__"):
            data = op.inputs()
        else:
            data = get_inputs_by_key(store, op.inputs)
    return data


def get_inputs_by_key(store, inputs_key):
    if isinstance(inputs_key, list):
        data = [store[key] for key in inputs_key]
    elif isinstance(inputs_key, tuple):
        data = tuple([store[key] for key in inputs_key])
    else:
        data = store[inputs_key]
    return data


def write_outputs_by_key(store, output, outputs_key):
    if isinstance(outputs_key, str):
        store[outputs_key] = output
    else:
        for key, data in zip(outputs_key, output):
            store[key] = data
    return store


def verify_ops(ops, class_name):
    inheritage = {"RecordWriter": NumpyOp, "Pipeline": TensorOp, "Network": TensorOp}
    inheritage_class = inheritage[class_name]
    if ops:
        assert ops[0].inputs, "must provide inputs for the operation '{}' in '{}'".format(
            type(ops[0]).__name__, class_name)
        assert ops[-1].outputs, "must provide outputs for the operation '{}' in '{}'".format(
            type(ops[-1]).__name__, class_name)
        inputs = ops[0].inputs
        for idx, op in enumerate(ops):
            assert isinstance(op,
                              inheritage_class), "operation '{}' in class '{}' doesn't have correct inheritage".format(
                                  type(op).__name__, class_name)
            if idx + 1 < len(ops) and ops[idx + 1].inputs:
                new_inputs = ops[idx + 1].inputs
                if new_inputs and new_inputs != inputs:
                    assert op.outputs, \
                        "must provide outputs for the operation '{}' in class '{}', otherwise the result will be lost"\
                        .format(type(op).__name__, class_name)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tqdm import trange

from fastestimator.pipeline.augmentation import Augmentation2D
from fastestimator.util.color_util import to_valid_rgb
from fastestimator.util.loader import PathLoader
from fastestimator.util.util import load_dict, load_image, decode_predictions
from fastestimator.util.vis_util import show_text, show_image, blur_image, fft_vars_to_whitened_im, rfft2d_freqs


@tf.function
def dot_loss(true, pred, cossim_pow):
    reduce_axis = tf.range(1, len(true.shape))  # Reductions should keep the batch axis (zero) separated
    # reduce_axis = tf.range(len(true.shape))  # TODO - it seems empirically like this doesn't matter. why???
    dot = tf.reduce_sum(true * pred, axis=reduce_axis)
    mag = tf.sqrt(tf.reduce_sum(true**2, axis=reduce_axis)) * tf.sqrt(tf.reduce_sum(pred**2, axis=reduce_axis))
    cossim = dot / (1e-6 + mag)
    cossim = tf.maximum(0.1, cossim)
    result = dot * cossim**cossim_pow
    # Negative value since we actually want to maximize this objective
    return -1. * result


@tf.function
def compute_gradients(model, aug, target_input, source_input, cossim_pow, decorrelate, sigmoid, fft, scale):
    aug.setup()
    transformed = tf.stack([aug.forward(target_input[i]) for i in range(target_input.shape[0])], axis=0)
    target = model(transformed)
    with tf.GradientTape() as tape:
        transformed = source_input
        if fft:
            transformed = fft_vars_to_whitened_im(transformed, scale, target_input.shape[2])
        transformed = to_valid_rgb(transformed, decorrelate=decorrelate, sigmoid=sigmoid)
        transformed = tf.stack([aug.forward(transformed[i]) for i in range(transformed.shape[0])], axis=0)
        layer_outputs = model(transformed)
        loss = dot_loss(target, layer_outputs, cossim_pow)
    return tape.gradient(loss, source_input)


def generate_variable_image(model_input, sd=0.01, fft=False, decay_power=1):
    sd = sd or 0.01
    if not fft:
        return tf.Variable(tf.random.normal(shape=model_input.shape, mean=tf.reduce_mean(model_input), stddev=sd),
                           trainable=True), None
    batch, h, w, ch = model_input.shape
    init_val_size = (2, batch, ch, h, w // 2 + 1 + (w % 2))  # zeroth index is for real/imaginary
    caricatures_freq = tf.Variable(tf.random.normal(shape=init_val_size, mean=0, stddev=sd))

    freqs = rfft2d_freqs(h, w)
    scale = 1.0 / np.maximum(freqs, 1.0 / max(w, h))**decay_power
    scale *= np.sqrt(w * h)
    big_kernel = tf.eye(num_rows=scale.shape[0], num_columns=scale.shape[0], batch_shape=[batch, ch], dtype=scale.dtype)
    scale = tf.matmul(big_kernel, scale)
    scale = tf.cast(scale, tf.complex64)
    return caricatures_freq, scale


def generate_caricatures(model, model_input, layer_id, n_steps=100, learning_rate=0.05, blur=1, cossim_pow=0.5, sd=0.01,
                         fft=True, decorrelate=True, sigmoid=True):
    small_model = tf.keras.models.Model(inputs=model.input, outputs=model.layers[layer_id].output)

    caricatures, scale = generate_variable_image(model_input, sd, fft, blur)

    optimizer = tf.optimizers.Adam(learning_rate=learning_rate, amsgrad=True)

    h = model_input.shape[1]
    w = model_input.shape[2]
    aug = Augmentation2D(rotation_range=5, zoom_range=0.05, shear_range=5, width_shift_range=5 / w,
                         height_shift_range=5 / h)
    aug.height = h
    aug.width = w

    for i in trange(n_steps, desc="Layer {}".format(layer_id)):
        gradients = compute_gradients(small_model, aug, model_input, caricatures, cossim_pow, decorrelate, sigmoid, fft,
                                      scale)
        # Optimizer call cannot be in tf.function b/c calling it on a newly instantiated object generates variables:
        # https://github.com/tensorflow/tensorflow/issues/27120
        optimizer.apply_gradients(zip([gradients], [caricatures]))
        if not fft:
            caricatures.assign(blur_image(caricatures, tf.constant(blur - blur * (i / n_steps))))

    if fft:
        caricatures = fft_vars_to_whitened_im(caricatures, scale, w)
    return to_valid_rgb(caricatures, decorrelate=decorrelate, sigmoid=sigmoid)


def caricature(model, model_input, layer_ids=None, decode_dictionary=None, save=False, save_path=".", n_steps=512,
               learning_rate=0.05, blur=1, cossim_pow=0.5, sd=0.01, fft=True, decorrelate=True, sigmoid=True):
    if layer_ids is None or len(layer_ids) == 0:
        layer_ids = [i for i in range(len(model.layers))]

    predictions = np.asarray(model(model_input))
    decoded = decode_predictions(predictions, top=3, dictionary=decode_dictionary)

    caricatures = [
        generate_caricatures(model, model_input, layer, n_steps=n_steps, learning_rate=learning_rate, blur=blur,
                             cossim_pow=cossim_pow, sd=sd, fft=fft, decorrelate=decorrelate, sigmoid=sigmoid)
        for layer in layer_ids
    ]

    num_rows = model_input.shape[0]
    num_cols = len(layer_ids) + 2
    dpi = 96.0

    box_width = max(220, model_input.shape[2])
    box_height = max(220, model_input.shape[1])
    fig, axs = plt.subplots(num_rows, num_cols, figsize=(num_cols * (box_width / dpi), num_rows * (box_height / dpi)),
                            dpi=dpi)

    if num_rows == 1:
        axs = [axs]  # axs object not wrapped if there's only one row
    for i in range(num_rows):
        show_text(axs[i][0], np.ones_like(model_input[i]), decoded[i], title="Predictions" if i == 0 else None)
        show_image(axs[i][1], model_input[i], title="Raw" if i == 0 else None)
        for j in range(len(layer_ids)):
            layer_id = layer_ids[j]
            layer_name = ": " + model.layers[layer_id].name
            show_image(axs[i][2 + j], caricatures[j][i],
                       title="Layer {}{}".format(layer_id, layer_name) if i == 0 else None)

    plt.subplots_adjust(top=0.95, bottom=0.01, left=0.01, right=0.99, hspace=0.03, wspace=0.03)
    if not save:
        plt.show()
    else:
        if save_path is None or save_path == "":
            save_path = "."
        os.makedirs(save_path, exist_ok=True)
        save_file = os.path.join(save_path, 'caricature.png')
        print("Saving to %s" % save_file)
        plt.savefig(save_file, dpi=300, bbox_inches="tight")


def load_and_caricature(model_path, input_paths, dictionary_path=None, save=False, save_dir=None, strip_alpha=False,
                        layer_ids=None, print_layers=False, n_steps=512, learning_rate=0.05, blur=1, cossim_pow=0.5,
                        sd=0.01, fft=True, decorrelate=True, sigmoid=True):
    model_dir = os.path.dirname(model_path)
    if save_dir is None:
        save_dir = model_dir
    network = keras.models.load_model(model_path)
    input_type = network.input.dtype
    input_shape = network.input.shape
    n_channels = 0 if len(input_shape) == 3 else input_shape[3]
    input_height = input_shape[1] or 224  # If the model doesn't specify width and height, just guess 224
    input_width = input_shape[2] or 224

    if print_layers:
        for idx, layer in enumerate(network.layers):
            print("{}: {} --- output shape: {}".format(idx, layer.name, layer.output_shape))
        return

    dic = load_dict(dictionary_path)
    if len(input_paths) == 1 and os.path.isdir(input_paths[0]):
        loader = PathLoader(input_paths[0])
        input_paths = [path[0] for path in loader.path_pairs]
    inputs = [load_image(input_paths[i], strip_alpha=strip_alpha, channels=n_channels) for i in range(len(input_paths))]
    tf_image = tf.stack([
        tf.image.resize_with_pad(tf.convert_to_tensor(im, dtype=input_type), input_height, input_width,
                                 method='lanczos3') for im in inputs
    ])
    tf_image = tf.clip_by_value(tf_image, -1, 1)

    caricature(network, tf_image, layer_ids=layer_ids, decode_dictionary=dic, save=save, save_path=save_dir,
               n_steps=n_steps, learning_rate=learning_rate, blur=blur, cossim_pow=cossim_pow, sd=sd, fft=fft,
               decorrelate=decorrelate, sigmoid=sigmoid)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import io
from unittest import TestCase
from unittest.mock import patch, MagicMock

from .parse_logs import *


class TestParser(TestCase):
    mock_good_file_path = '/good/file1.txt'
    mock_good_file_contents = 'FastEstimator: Model artifact will be saved in /opt/ml/model\nFastEstimator: Found 10092 examples for training and 2523 for validation in /opt/ml/input/data/training\nFastEstimator-Train:step: 0; train_loss: 12.027558; lr: 0.000200; example/sec: 0.000000;\nFastEstimator-Train:step: 100; train_loss: 2.565781; lr: 0.000200; example/sec: 44.738688;\nFastEstimator-Train:step: 200; train_loss: 0.824913; lr: 0.000200; example/sec: 45.086421;\nFastEstimator-Eval:step: 281; val_loss: 0.725258; min_val_loss: 0.725258; since_best: 0; val_mask_raw_loss: -0.007752625846316934; val_image_labels_loss: 0.7330105359355609; val_mask_raw_conditionalDice: 0.007752625846316934; val_image_labels_my_binary_accuracy: 0.5194444588075081;\nFastEstimator-Train:step: 300; train_loss: 0.561318; lr: 0.000200; example/sec: 44.689092;\nFastEstimator-Train:step: 400; train_loss: 0.427389; lr: 0.000200; example/sec: 44.799198;\nFastEstimator-Train:step: 500; train_loss: 0.528405; lr: 0.000200; example/sec: 44.523727;\nFastEstimator-Eval:step: 562; val_loss: 4.125795; min_val_loss: 0.725258; since_best: 1; val_mask_raw_loss: -0.15434368319272632; val_image_labels_loss: 4.280138591821823; val_mask_raw_conditionalDice: 0.15434368319272632; val_image_labels_my_binary_accuracy: 0.5662698552840286;\nFastEstimator-Train:step: 600; train_loss: 0.686736; lr: 0.000200; example/sec: 45.055799;'
    mock_good_parsed_result = {
        "train_loss": [[0, 12.027558], [100, 2.565781], [200, 0.824913], [300, 0.561318], [400, 0.427389],
                       [500, 0.528405], [600, 0.686736]],
        "lr": [[0, 0.0002], [100, 0.0002], [200, 0.0002], [300, 0.0002], [400, 0.0002], [500, 0.0002], [600, 0.0002]],
        "example/sec": [[0, 0.0], [100, 44.738688], [200, 45.086421], [300, 44.689092], [400, 44.799198],
                        [500, 44.523727], [600, 45.055799]],
        "val_loss": [[281, 0.725258], [562, 4.125795]],
        "min_val_loss": [[281, 0.725258], [562, 0.725258]],
        "since_best": [[281, 0.0], [562, 1.0]],
        "val_mask_raw_loss": [[281, -0.007752625846316934], [562, -0.15434368319272632]],
        "val_image_labels_loss": [[281, 0.7330105359355609], [562, 4.280138591821823]],
        "val_mask_raw_conditionalDice": [[281, 0.007752625846316934], [562, 0.15434368319272632]],
        "val_image_labels_my_binary_accuracy": [[281, 0.5194444588075081], [562, 0.5662698552840286]]
    }
    mock_good_file_path_2 = '/good/file2.txt'
    mock_good_file_contents_2 = 'FastEstimator: Model artifact will be saved in /opt/ml/model\nFastEstimator: Found 10092 examples for training and 2523 for validation in /opt/ml/input/data/training\nFastEstimator-Train:step: 0; train_loss: 3.038919; lr: 0.000200; example/sec: 0.000000;\nFastEstimator-Train:step: 100; train_loss: 0.887208; lr: 0.000200; example/sec: 44.704689;\nFastEstimator-Train:step: 200; train_loss: 0.541103; lr: 0.000200; example/sec: 44.553714;\nFastEstimator-Eval:step: 281; val_loss: 3.122921; min_val_loss: 3.122921; since_best: 0; val_mask_raw_loss: -0.08894996416088866; val_image_labels_loss: 3.2118707436654303; val_mask_raw_conditionalDice: 0.08894996416088866; val_image_labels_my_binary_accuracy: 0.5757936653577619;\nFastEstimator-Train:step: 300; train_loss: 1.814015; lr: 0.000200; example/sec: 45.014834;\nFastEstimator-Train:step: 400; train_loss: 0.485160; lr: 0.000200; example/sec: 44.368620;\nFastEstimator-Train:step: 500; train_loss: 0.375250; lr: 0.000200; example/sec: 44.782828;\nFastEstimator-Eval:step: 562; val_loss: 1.244561; min_val_loss: 1.244561; since_best: 0; val_mask_raw_loss: -0.13716965285970623; val_image_labels_loss: 1.3817303948311341; val_mask_raw_conditionalDice: 0.13716965285970623; val_image_labels_my_binary_accuracy: 0.624603189734949;\nFastEstimator-Train:step: 600; train_loss: 0.481111; lr: 0.000200; example/sec: 44.513710;\nFastEstimator-Train:step: 700; train_loss: 0.308867; lr: 0.000200; example/sec: 44.525426;\nFastEstimator-Train:step: 800; train_loss: 0.343235; lr: 0.000200; example/sec: 44.179674;\nFastEstimator-Eval:step: 843; val_loss: 1.007462; min_val_loss: 1.007462; since_best: 0; val_mask_raw_loss: -0.12723745336374526; val_image_labels_loss: 1.1346999171293444; val_mask_raw_conditionalDice: 0.12723745336374526; val_image_labels_my_binary_accuracy: 0.5924603325625261;\nFastEstimator-Train:step: 900; train_loss: 0.327094; lr: 0.000200; example/sec: 44.503702;\nFastEstimator-Train:step: 1000; train_loss: 0.555805; lr: 0.000200; example/sec: 44.428722;\nFastEstimator-Train:step: 1100; train_loss: 2.290446; lr: 0.000200; example/sec: 44.082693;\nFastEstimator-Eval:step: 1124; val_loss: 0.877936; min_val_loss: 0.877936; since_best: 0; val_mask_raw_loss: -0.21344286083541292; val_image_labels_loss: 1.0913789242506027; val_mask_raw_conditionalDice: 0.21344286083541292; val_image_labels_my_binary_accuracy: 0.6357143003907468;\nFastEstimator-Train:step: 1200; train_loss: 0.727391; lr: 0.000200; example/sec: 44.769331;\nFastEstimator-Train:step: 1300; train_loss: 1.556982; lr: 0.000200; example/sec: 44.300033;\nFastEstimator-Train:step: 1400; train_loss: 0.563514; lr: 0.000200; example/sec: 44.115453;\nFastEstimator-Eval:step: 1405; val_loss: 1.220150; min_val_loss: 0.877936; since_best: 1; val_mask_raw_loss: -0.22651713909775328; val_image_labels_loss: 1.4466672077775002; val_mask_raw_conditionalDice: 0.22651713909775328; val_image_labels_my_binary_accuracy: 0.6488095390299956;\nFastEstimator-Train:step: 1500; train_loss: 0.704434; lr: 0.000200; example/sec: 44.554155;'
    mock_good_parsed_result_2 = {
        "train_loss": [[0, 3.038919], [100, 0.887208], [200, 0.541103], [300, 1.814015], [400, 0.48516], [500, 0.37525],
                       [600, 0.481111], [700, 0.308867], [800, 0.343235], [900, 0.327094], [1000, 0.555805],
                       [1100, 2.290446], [1200, 0.727391], [1300, 1.556982], [1400, 0.563514], [1500, 0.704434]],
        "lr": [[0, 0.0002], [100, 0.0002], [200, 0.0002], [300, 0.0002], [400, 0.0002], [500, 0.0002], [600, 0.0002],
               [700, 0.0002], [800, 0.0002], [900, 0.0002], [1000, 0.0002], [1100, 0.0002], [1200, 0.0002],
               [1300, 0.0002], [1400, 0.0002], [1500, 0.0002]],
        "example/sec": [[0, 0.0], [100, 44.704689], [200, 44.553714], [300, 45.014834], [400, 44.36862],
                        [500, 44.782828], [600, 44.51371], [700, 44.525426], [800, 44.179674], [900, 44.503702],
                        [1000, 44.428722], [1100, 44.082693], [1200, 44.769331], [1300, 44.300033], [1400, 44.115453],
                        [1500, 44.554155]],
        "val_loss": [[281, 3.122921], [562, 1.244561], [843, 1.007462], [1124, 0.877936], [1405, 1.22015]],
        "min_val_loss": [[281, 3.122921], [562, 1.244561], [843, 1.007462], [1124, 0.877936], [1405, 0.877936]],
        "since_best": [[281, 0.0], [562, 0.0], [843, 0.0], [1124, 0.0], [1405, 1.0]],
        "val_mask_raw_loss": [[281, -0.08894996416088866], [562, -0.13716965285970623], [843, -0.12723745336374526],
                              [1124, -0.21344286083541292], [1405, -0.22651713909775328]],
        "val_image_labels_loss": [[281, 3.2118707436654303], [562, 1.3817303948311341], [843, 1.1346999171293444],
                                  [1124, 1.0913789242506027], [1405, 1.4466672077775002]],
        "val_mask_raw_conditionalDice": [[281, 0.08894996416088866], [562, 0.13716965285970623],
                                         [843, 0.12723745336374526], [1124, 0.21344286083541292],
                                         [1405, 0.22651713909775328]],
        "val_image_labels_my_binary_accuracy": [[281, 0.5757936653577619], [562, 0.624603189734949],
                                                [843, 0.5924603325625261], [1124, 0.6357143003907468],
                                                [1405, 0.6488095390299956]]
    }
    mock_empty_file_path = '/bad/empty.txt'
    mock_empty_file_contents = ''
    mock_fake_file_path = '/bad/fake.txt'
    mock_bad_file_path_missing_step = '/bad/file.txt'
    mock_bad_file_contents_missing_step = 'FastEstimator: Model artifact will be saved in /opt/ml/model\nFastEstimator: Found 10092 examples for training and 2523 for validation in /opt/ml/input/data/training\nFastEstimator-Train:step: 0; train_loss: 12.027558; lr: 0.000200; example/sec: 0.000000;\nFastEstimator-Train:step: 100; train_loss: 2.565781; lr: 0.000200; example/sec: 44.738688;\nFastEstimator-Train:step: 200; train_loss: 0.824913; lr: 0.000200; example/sec: 45.086421;\nFastEstimator-Eval:step: 281; val_loss: 0.725258; min_val_loss: 0.725258; since_best: 0; val_mask_raw_loss: -0.007752625846316934; val_image_labels_loss: 0.7330105359355609; val_mask_raw_conditionalDice: 0.007752625846316934; val_image_labels_my_binary_accuracy: 0.5194444588075081;\nFastEstimator-Train:train_loss: 0.561318; lr: 0.000200; example/sec: 44.689092;\nFastEstimator-Train:step: 400; train_loss: 0.427389; lr: 0.000200; example/sec: 44.799198;\nFastEstimator-Train:step: 500; train_loss: 0.528405; lr: 0.000200; example/sec: 44.523727;\nFastEstimator-Eval:step: 562; val_loss: 4.125795; min_val_loss: 0.725258; since_best: 1; val_mask_raw_loss: -0.15434368319272632; val_image_labels_loss: 4.280138591821823; val_mask_raw_conditionalDice: 0.15434368319272632; val_image_labels_my_binary_accuracy: 0.5662698552840286;\nFastEstimator-Train:step: 600; train_loss: 0.686736; lr: 0.000200; example/sec: 45.055799;'

    @staticmethod
    def load_mock_file(path):
        if TestParser.mock_good_file_path == path:
            return io.StringIO(TestParser.mock_good_file_contents)
        elif TestParser.mock_good_file_path_2 == path:
            return io.StringIO(TestParser.mock_good_file_contents_2)
        elif TestParser.mock_bad_file_path_missing_step == path:
            return io.StringIO(TestParser.mock_bad_file_contents_missing_step)
        elif TestParser.mock_empty_file_path == path:
            return io.StringIO(TestParser.mock_empty_file_contents)
        else:
            raise FileNotFoundError("[Errno 2] No such file or directory: '%s'" % path)

    # -------------------------------------------------------------------------------------------------------- #
    # ------------------------------------------------ Parser ------------------------------------------------ #
    # -------------------------------------------------------------------------------------------------------- #
    def test_parser_success(self):
        expected = TestParser.mock_good_parsed_result
        # m = mock_open(read_data=mock_good_file_contents)  # Could use this instead of magic mock
        m = MagicMock(side_effect=TestParser.load_mock_file)
        with patch('fastestimator.visualization.parse_logs.open', m):
            actual = parse_file(TestParser.mock_good_file_path)
        self.assertDictEqual(actual, expected)

    def test_parser_success_2(self):
        expected = TestParser.mock_good_parsed_result_2
        m = MagicMock(side_effect=TestParser.load_mock_file)
        with patch('fastestimator.visualization.parse_logs.open', m):
            actual = parse_file(TestParser.mock_good_file_path_2)
        self.assertDictEqual(actual, expected)

    def test_parser_empty_file(self):
        expected = {}
        m = MagicMock(side_effect=TestParser.load_mock_file)
        with patch('fastestimator.visualization.parse_logs.open', m):
            actual = parse_file(TestParser.mock_empty_file_path)
        self.assertDictEqual(actual, expected)

    def test_parser_file_does_not_exist(self):
        m = MagicMock(side_effect=TestParser.load_mock_file)
        with patch('fastestimator.visualization.parse_logs.open', m):
            self.assertRaises(FileNotFoundError, parse_file, TestParser.mock_fake_file_path)
        m.assert_called_once_with(TestParser.mock_fake_file_path)

    def test_parser_fail_missing_step(self):
        m = MagicMock(side_effect=TestParser.load_mock_file)
        with patch('fastestimator.visualization.parse_logs.open', m):
            self.assertRaises(AssertionError, parse_file, TestParser.mock_bad_file_path_missing_step)
        m.assert_called_once_with(TestParser.mock_bad_file_path_missing_step)

    # -------------------------------------------------------------------------------------------------------- #
    # --------------------------------------------- Parse Files ---------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_parse_files(self):
        expected = {'file1': TestParser.mock_good_parsed_result, 'file2': TestParser.mock_good_parsed_result_2}

        parse_mock = MagicMock(side_effect=TestParser.load_mock_file)
        graph_mock = MagicMock()
        with patch('fastestimator.visualization.parse_logs.open', parse_mock):
            with patch('fastestimator.visualization.parse_logs.graph_metrics', graph_mock):
                parse_files([TestParser.mock_good_file_path, TestParser.mock_good_file_path_2], log_extension='.txt',
                            smooth_factor=1.5, save=True, save_path='/out/', ignore_metrics=None, share_legend=False,
                            pretty_names=True)
        graph_mock.assert_called_once_with(expected, 1.5, True, '/out/', False, True)

    def test_parse_files_no_save_dir(self):
        expected = {'file1': TestParser.mock_good_parsed_result, 'file2': TestParser.mock_good_parsed_result_2}

        parse_mock = MagicMock(side_effect=TestParser.load_mock_file)
        graph_mock = MagicMock()
        with patch('fastestimator.visualization.parse_logs.open', parse_mock):
            with patch('fastestimator.visualization.parse_logs.graph_metrics', graph_mock):
                parse_files([TestParser.mock_good_file_path, TestParser.mock_good_file_path_2], log_extension='.txt',
                            smooth_factor=1.5, save=True, save_path=None, ignore_metrics=None, share_legend=False,
                            pretty_names=True)
        graph_mock.assert_called_once_with(expected, 1.5, True, TestParser.mock_good_file_path, False, True)

    def test_parse_files_no_files(self):
        self.assertRaises(AssertionError, parse_files, [])

    # -------------------------------------------------------------------------------------------------------- #
    # --------------------------------------------- Parse Folder --------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_parse_folder(self):
        mock_dir = '/good/'
        parse_mock = MagicMock()
        is_dir_mock = MagicMock(return_value=True)
        walk_mock = MagicMock(return_value=[('/good', (), ('file1.log', 'file2.log', 'file3.txt', 'file4'))])
        with patch('fastestimator.visualization.parse_logs.parse_files',
                   parse_mock), patch('fastestimator.util.loader.os.path.isdir',
                                      is_dir_mock), patch('fastestimator.util.loader.os.walk', walk_mock):
            parse_folder(mock_dir, log_extension='.log', smooth_factor=2, save=True, save_path=None,
                         ignore_metrics=['lr'])
        is_dir_mock.assert_called_once_with(mock_dir)
        walk_mock.assert_called_once_with(mock_dir)
        parse_mock.assert_called_once_with(['/good/file1.log', '/good/file2.log'], '.log', 2, True, None, ['lr'], True,
                                           False)

    def test_parse_folder_bad_dir(self):
        mock_dir = '/bad/'
        parse_mock = MagicMock()
        is_dir_mock = MagicMock(return_value=False)
        walk_mock = MagicMock(return_value=[('/good', (), ('file1.log', 'file2.log', 'file3.txt', 'file4'))])
        with patch('fastestimator.visualization.parse_logs.parse_files',
                   parse_mock), patch('fastestimator.util.loader.os.path.isdir',
                                      is_dir_mock), patch('fastestimator.util.loader.os.walk', walk_mock):
            self.assertRaises(AssertionError, parse_folder, mock_dir)
        is_dir_mock.assert_called_once_with(mock_dir)

    def test_parse_folder_empty_dir(self):
        mock_dir = '/good/'
        parse_mock = MagicMock()
        is_dir_mock = MagicMock(return_value=True)
        walk_mock = MagicMock(return_value=[('/good', (), ())])
        with patch('fastestimator.visualization.parse_logs.parse_files',
                   parse_mock), patch('fastestimator.util.loader.os.path.isdir',
                                      is_dir_mock), patch('fastestimator.util.loader.os.walk', walk_mock):
            parse_folder(mock_dir, log_extension='.log', smooth_factor=2, save=True, save_path=None,
                         ignore_metrics=['lr'])
        is_dir_mock.assert_called_once_with(mock_dir)
        walk_mock.assert_called_once_with(mock_dir)
        parse_mock.assert_called_once_with([], '.log', 2, True, None, ['lr'], True, False)

    # -------------------------------------------------------------------------------------------------------- #
    # ----------------------------------------------- Graphing ----------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_graph_metrics_save(self):
        metrics = {'trial1': TestParser.mock_good_parsed_result, 'trial2': TestParser.mock_good_parsed_result_2}
        fig_mock = MagicMock()
        dir_mock = MagicMock()
        with patch('fastestimator.visualization.parse_logs.plt.savefig',
                   fig_mock), patch('fastestimator.visualization.parse_logs.os.makedirs', dir_mock):
            graph_metrics(metrics, 0, True, '/good/folder/trial1.txt', True, False)
        dir_mock.assert_called_once_with('/good/folder', exist_ok=True)
        fig_mock.assert_called_once_with('/good/folder/parse_logs.png', dpi=300)

    def test_graph_metrics_display(self):
        metrics = {'trial1': TestParser.mock_good_parsed_result, 'trial2': TestParser.mock_good_parsed_result_2}
        fig_mock = MagicMock()
        with patch('fastestimator.visualization.parse_logs.plt.show', fig_mock):
            graph_metrics(metrics, 2, False, None, True, False)
        # fig_mock.assert_called_once()  # This method doesn't exist in python 3.5 :(
        assert len(fig_mock.mock_calls) == 1, "Execution failed before figure fall invoked"

    def test_graph_metrics_display_2_metrics(self):
        blacklist = [
            'lr', 'min_val_loss', 'since_best', 'train_loss', 'val_mask_raw_loss', 'val_mask_raw_conditionalDice',
            'val_loss', 'val_image_labels_my_binary_accuracy'
        ]
        reduced_parse_1 = copy.deepcopy(TestParser.mock_good_parsed_result)
        reduced_parse_2 = copy.deepcopy(TestParser.mock_good_parsed_result_2)
        remove_blacklist_keys(reduced_parse_1, blacklist)
        remove_blacklist_keys(reduced_parse_2, blacklist)
        metrics = {'trial1': reduced_parse_1, 'trial2': reduced_parse_2}
        fig_mock = MagicMock()
        with patch('fastestimator.visualization.parse_logs.plt.show', fig_mock):
            graph_metrics(metrics, 2, False, None, True, False)
        assert len(fig_mock.mock_calls) == 1, "Execution failed before figure fall invoked"

    def test_graph_metrics_display_1_metrics(self):
        blacklist = [
            'lr', 'min_val_loss', 'since_best', 'train_loss', 'val_mask_raw_loss', 'val_mask_raw_conditionalDice',
            'val_loss', 'val_image_labels_my_binary_accuracy', 'val_image_labels_loss'
        ]
        reduced_parse_1 = copy.deepcopy(TestParser.mock_good_parsed_result)
        reduced_parse_2 = copy.deepcopy(TestParser.mock_good_parsed_result_2)
        remove_blacklist_keys(reduced_parse_1, blacklist)
        remove_blacklist_keys(reduced_parse_2, blacklist)
        metrics = {'trial1': reduced_parse_1, 'trial2': reduced_parse_2}
        fig_mock = MagicMock()
        with patch('fastestimator.visualization.parse_logs.plt.show', fig_mock):
            graph_metrics(metrics, 2, False, None, True, False)
        assert len(fig_mock.mock_calls) == 1, "Execution failed before figure fall invoked"

    def test_graph_metrics_display_0_metrics(self):
        blacklist = [
            'lr', 'min_val_loss', 'since_best', 'train_loss', 'val_mask_raw_loss', 'val_mask_raw_conditionalDice',
            'val_loss', 'val_image_labels_my_binary_accuracy', 'val_image_labels_loss', 'example/sec'
        ]
        reduced_parse_1 = copy.deepcopy(TestParser.mock_good_parsed_result)
        reduced_parse_2 = copy.deepcopy(TestParser.mock_good_parsed_result_2)
        remove_blacklist_keys(reduced_parse_1, blacklist)
        remove_blacklist_keys(reduced_parse_2, blacklist)
        metrics = {'trial1': reduced_parse_1, 'trial2': reduced_parse_2}
        self.assertRaises(AssertionError, graph_metrics, metrics, 2, False, None, True, False)

    def test_graph_metrics_save_no_files(self):
        metrics = {}
        self.assertRaises(AssertionError, graph_metrics, metrics, 2, False, None, True, False)

    def test_graph_metrics_bad_smooth(self):
        metrics = {'trial1': TestParser.mock_good_parsed_result, 'trial2': TestParser.mock_good_parsed_result_2}
        self.assertRaises(AssertionError, graph_metrics, metrics, -1, False, None, True, False)



import os
import tempfile
import zipfile
import wget

import pandas as pd
from glob import glob

NUM_IMG_FILES = 82783


def _download_and_extract(url, filename, path):
    if not os.path.exists(os.path.join(path, filename)):
        print("Downloading data to {}".format(path))
        wget.download(url, path)

    with zipfile.ZipFile(os.path.join(path, filename), 'r') as zip_file:
        print("Extracting {}".format(os.path.join(path, filename)))
        zip_file.extractall(path)


def _create_csv(data_path):
    filenames = glob(os.path.join(data_path, '*.jpg'))
    csv = None
    if len(filenames) == NUM_IMG_FILES:
        csv = pd.DataFrame()
        rel_filenames = [f.replace(data_path, '.') for f in filenames]
        csv["image"] = rel_filenames
    else:
        print("One or more images are missing")
    return csv


def load_data(path=None):
    url = {"train": "http://images.cocodataset.org/zips/train2014.zip"}
    if path is None:
        path = os.path.join(tempfile.gettempdir(), ".fe", "MSCOCO2014")
    path = os.path.abspath(path)

    os.makedirs(path, exist_ok=True)
    train_csv_path = os.path.join(path, "coco_train.csv")
    train_data_dir = os.path.join(path, "train2014")
    if not (os.path.exists(train_data_dir)):
        _download_and_extract(url["train"], "train2014.zip", path)
        train_csv = _create_csv(train_data_dir)
        train_csv.to_csv(train_csv_path, index=False)
    else:
        if not (len(glob(os.path.join(train_data_dir, '*.jpg'))) == NUM_IMG_FILES):
            print("One or more images are missing")
            _download_and_extract(url["train"], "train2014.zip", path)
            train_csv = _create_csv(train_data_dir)
            train_csv.to_csv(train_csv_path, index=False)
        else:
            if not (os.path.exists(train_csv_path)):
                train_csv = _create_csv(train_data_dir)
                train_csv.to_csv(train_csv_path, index=False)
            else:
                print("reusing existing dataset")
    return train_csv_path, train_data_dir



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from unittest import TestCase

from tensorflow.python.client import device_lib

from .util import (get_num_devices, parse_cli_to_dictionary, parse_string_to_python, prettify_metric_name,
                   remove_blacklist_keys, strip_suffix)


class TestUtil(TestCase):
    mock_good_parsed_result = {
        "train_loss": [[0, 12.027558], [100, 2.565781], [200, 0.824913], [300, 0.561318], [400, 0.427389],
                       [500, 0.528405], [600, 0.686736]],
        "lr": [[0, 0.0002], [100, 0.0002], [200, 0.0002], [300, 0.0002], [400, 0.0002], [500, 0.0002], [600, 0.0002]],
        "example/sec": [[0, 0.0], [100, 44.738688], [200, 45.086421], [300, 44.689092], [400, 44.799198],
                        [500, 44.523727], [600, 45.055799]],
        "val_loss": [[281, 0.725258], [562, 4.125795]],
        "min_val_loss": [[281, 0.725258], [562, 0.725258]],
        "since_best": [[281, 0.0], [562, 1.0]],
        "val_mask_raw_loss": [[281, -0.007752625846316934], [562, -0.15434368319272632]],
        "val_image_labels_loss": [[281, 0.7330105359355609], [562, 4.280138591821823]],
        "val_mask_raw_conditionalDice": [[281, 0.007752625846316934], [562, 0.15434368319272632]],
        "val_image_labels_my_binary_accuracy": [[281, 0.5194444588075081], [562, 0.5662698552840286]]
    }

    # -------------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------- GPU Count ----------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #

    def test_get_num_devices(self):
        try:
            result = subprocess.run(['nvidia-smi', '-q'], stdout=subprocess.PIPE).stdout.decode('utf-8')
            lines = [line.split() for line in result.splitlines() if line.startswith("Attached GPUs")]
            devices = int(lines[0][-1])
        except:
            devices = 1
        assert devices == get_num_devices()

    # -------------------------------------------------------------------------------------------------------- #
    # ------------------------------------------- KEY Blacklisting ------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_remove_blacklist_keys_success(self):
        expected = {
            "train_loss": [[0, 12.027558], [100, 2.565781], [200, 0.824913], [300, 0.561318], [400, 0.427389],
                           [500, 0.528405], [600, 0.686736]],
            "lr": [[0, 0.0002], [100, 0.0002], [200, 0.0002], [300, 0.0002], [400, 0.0002], [500, 0.0002],
                   [600, 0.0002]],
            "val_image_labels_loss": [[281, 0.7330105359355609], [562, 4.280138591821823]]
        }
        blacklist = [
            "val_loss", "since_best", "example/sec", "min_val_loss", "val_mask_raw_loss",
            "val_mask_raw_conditionalDice", "val_image_labels_my_binary_accuracy"
        ]
        actual = copy.deepcopy(self.mock_good_parsed_result)
        remove_blacklist_keys(actual, blacklist)
        self.assertDictEqual(actual, expected)

    def test_remove_blacklist_keys_none(self):
        expected = self.mock_good_parsed_result
        blacklist = None
        actual = copy.deepcopy(self.mock_good_parsed_result)
        remove_blacklist_keys(actual, blacklist)
        self.assertDictEqual(actual, expected)

    def test_remove_blacklist_keys_empty_list(self):
        expected = self.mock_good_parsed_result
        blacklist = []
        actual = copy.deepcopy(self.mock_good_parsed_result)
        remove_blacklist_keys(actual, blacklist)
        self.assertDictEqual(actual, expected)

    def test_remove_blacklist_keys_empty_set(self):
        expected = self.mock_good_parsed_result
        blacklist = {}
        actual = copy.deepcopy(self.mock_good_parsed_result)
        remove_blacklist_keys(actual, blacklist)
        self.assertDictEqual(actual, expected)

    def test_remove_blacklist_keys_empty(self):
        expected = {}
        blacklist = ["FAKE_KEY"]
        actual = {}
        remove_blacklist_keys(actual, blacklist)
        self.assertDictEqual(actual, expected)

    def test_remove_blacklist_keys_missing(self):
        expected = {
            "train_loss": [[0, 12.027558], [100, 2.565781], [200, 0.824913], [300, 0.561318], [400, 0.427389],
                           [500, 0.528405], [600, 0.686736]],
            "lr": [[0, 0.0002], [100, 0.0002], [200, 0.0002], [300, 0.0002], [400, 0.0002], [500, 0.0002],
                   [600, 0.0002]],
            "val_image_labels_loss": [[281, 0.7330105359355609], [562, 4.280138591821823]]
        }
        blacklist = [
            "val_loss", "since_best", "example/sec", "min_val_loss", "val_mask_raw_loss",
            "val_mask_raw_conditionalDice", "val_image_labels_my_binary_accuracy", "FAKE_KEY"
        ]
        actual = copy.deepcopy(self.mock_good_parsed_result)
        remove_blacklist_keys(actual, blacklist)
        self.assertDictEqual(actual, expected)

    # -------------------------------------------------------------------------------------------------------- #
    # ------------------------------------------------ Suffix ------------------------------------------------ #
    # -------------------------------------------------------------------------------------------------------- #
    def test_strip_suffix(self):
        base = "ImageNet.txt"
        suffix = ".txt"
        expected = "ImageNet"
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_empty(self):
        base = "ImageNet.txt"
        suffix = ""
        expected = "ImageNet.txt"
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_none(self):
        base = "ImageNet.txt"
        suffix = None
        expected = "ImageNet.txt"
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_full(self):
        base = "ImageNet.txt"
        suffix = "ImageNet.txt"
        expected = ""
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_wrong_suffix(self):
        base = "ImageNet.txt"
        suffix = ".tzt"
        expected = "ImageNet.txt"
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_super_suffix(self):
        base = "ImageNet.txt"
        suffix = "MImageNet.txt"
        expected = "ImageNet.txt"
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_empty_base(self):
        base = ""
        suffix = ".txt"
        expected = ""
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    def test_strip_suffix_no_base(self):
        base = None
        suffix = ".txt"
        expected = None
        actual = strip_suffix(base, suffix)
        self.assertEqual(actual, expected)

    # -------------------------------------------------------------------------------------------------------- #
    # ----------------------------------------------- Prettify ----------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_prettify_metric_name(self):
        base = "val_mask_raw_conditionalDice"
        expected = "Val Mask Raw Conditional Dice"
        actual = prettify_metric_name(base)
        self.assertEqual(actual, expected)

    # -------------------------------------------------------------------------------------------------------- #
    # --------------------------------------------- Parse String --------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_parse_string_to_python_none(self):
        input_string = None
        expected = ""
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python(self):
        input_string = ""
        expected = ""
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_array(self):
        input_string = '[]'
        expected = []
        actual = parse_string_to_python(input_string)
        self.assertListEqual(actual, expected)

    def test_parse_string_to_python_tuple(self):
        input_string = '()'
        expected = ()
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_dict(self):
        input_string = '{}'
        expected = {}
        actual = parse_string_to_python(input_string)
        self.assertDictEqual(actual, expected)

    def test_parse_string_to_python_true(self):
        input_string = "true"
        expected = True
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_true2(self):
        input_string = "True"
        expected = True
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_false(self):
        input_string = "false"
        expected = False
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_false2(self):
        input_string = "False"
        expected = False
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_boolean_array(self):
        input_string = '[true, false, true, false]'
        expected = [True, False, True, False]
        actual = parse_string_to_python(input_string)
        self.assertListEqual(actual, expected)

    def test_parse_string_to_python_boolean_array2(self):
        input_string = '[True, False, True, False]'
        expected = [True, False, True, False]
        actual = parse_string_to_python(input_string)
        self.assertListEqual(actual, expected)

    def test_parse_string_to_python_boolean_tuple(self):
        input_string = '(True, False, True, False)'
        expected = (True, False, True, False)
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_int(self):
        input_string = "7"
        expected = 7
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_int_array(self):
        input_string = '[0, -2, 4, 8]'
        expected = [0, -2, 4, 8]
        actual = parse_string_to_python(input_string)
        self.assertListEqual(actual, expected)

    def test_parse_string_to_python_int_tuple(self):
        input_string = '(0, -2, 4, 8)'
        expected = (0, -2, 4, 8)
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_float(self):
        input_string = "7.5"
        expected = 7.5
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_float_array(self):
        input_string = '[0.5, -2.1, 4, 8.89]'
        expected = [0.5, -2.1, 4, 8.89]
        actual = parse_string_to_python(input_string)
        self.assertListEqual(actual, expected)

    def test_parse_string_to_python_float_tuple(self):
        input_string = '(0.5, -2.1, 4, 8.89)'
        expected = (0.5, -2.1, 4, 8.89)
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_string(self):
        input_string = "random string"
        expected = "random string"
        actual = parse_string_to_python(input_string)
        self.assertEqual(actual, expected)

    def test_parse_string_to_python_string_array(self):
        input_string = '["random", "string"]'
        expected = ['random', 'string']
        actual = parse_string_to_python(input_string)
        self.assertListEqual(actual, expected)

    def test_parse_string_to_python_string_tuple(self):
        input_string = '("random", "string")'
        expected = ('random', 'string')
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_string_tuple2(self):
        input_string = "('random', 'string')"
        expected = ('random', 'string')
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_nested(self):
        input_string = '("random", ["string1", 10], (True, 7.5, "string2"))'
        expected = ('random', ["string1", 10], (True, 7.5, "string2"))
        actual = parse_string_to_python(input_string)
        self.assertTupleEqual(actual, expected)

    def test_parse_string_to_python_string_dict(self):
        input_string = '{"key1":"val1","key2":"val2"}'
        expected = {"key1": "val1", "key2": "val2"}
        actual = parse_string_to_python(input_string)
        self.assertDictEqual(actual, expected)

    # -------------------------------------------------------------------------------------------------------- #
    # ----------------------------------------------- Parse CLI ---------------------------------------------- #
    # -------------------------------------------------------------------------------------------------------- #
    def test_parse_cli_to_dictionary_none(self):
        input_list = None
        expected = {}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)

    def test_parse_cli_to_dictionary(self):
        input_list = []
        expected = {}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)

    def test_parse_cli_to_dictionary_no_key(self):
        input_list = ["thing1", "thing2", "True", "(0,", "1)"]
        expected = {}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)

    def test_parse_cli_to_dictionary_one_key_string(self):
        input_list = ["thing1", "--key1", "True", "(0,", "1)"]
        expected = {"key1": 'True(0,1)'}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)

    def test_parse_cli_to_dictionary_one_key_string2(self):
        input_list = ["--key1", "True"]
        expected = {"key1": True}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)

    def test_parse_cli_to_dictionary_one_key_tuple(self):
        input_list = ["thing1", "--key1", "(0,", "1)"]
        expected = {"key1": (0, 1)}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)

    def test_parse_cli_to_dictionary_two_key_tuple(self):
        input_list = ["--key1", "(0,", "1)", "--key2", "[True,", "False,", "'args']"]
        expected = {"key1": (0, 1), "key2": [True, False, 'args']}
        actual = parse_cli_to_dictionary(input_list)
        self.assertDictEqual(actual, expected)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import re
import string
import sys
import time
from ast import literal_eval
from contextlib import ContextDecorator
from itertools import chain

import numpy as np
import PIL
# noinspection PyPackageRequirements
import tensorflow as tf
from tensorflow.python.client import device_lib


def load_image(file_path, strip_alpha=False, channels=3):
    """
    Args:
        file_path: The path to an image file
        strip_alpha: True to convert an RGBA image to RGB
        channels: How many channels should the image have (0,1,3)
    Returns:
        The image loaded into memory and scaled to a range of [-1, 1]
    """
    # noinspection PyUnresolvedReferences
    im = PIL.Image.open(file_path)
    if strip_alpha and im.mode == "RGBA":
        # noinspection PyUnresolvedReferences
        background = PIL.Image.new("RGB", im.size, (0, 0, 0))
        background.paste(im, mask=im.split()[3])
        im = background
    if channels == 0 or channels == 1:
        im = im.convert("L")
    if channels == 3:
        im = im.convert("RGB")
    im = np.asarray(im) / 127.5 - 1.0
    if channels == 1:
        im = np.reshape(im, (im.shape[0], im.shape[1], 1))
    return im


def load_dict(dict_path, array_key=False):
    """
    Args:
        dict_path: The path to a json dictionary
        array_key: If true the parser will consider the first element in a sublist at the key {_:[K,...,V],...}.
                    Otherwise it will parse as {K:V,...} or {K:[...,V],...}
    Returns:
        A dictionary corresponding to the info from the file. If the file was formatted with arrays as the values for a
         key, the last element of the array is used as the value for the key in the parsed dictionary
    """
    parsed = None
    if dict_path is not None:
        with open(dict_path) as f:
            parsed = json.load(f)
            for key in list(parsed.keys()):
                entry = parsed[key]
                if type(entry) == list:
                    if array_key:
                        val = parsed.pop(key)
                        parsed[val[0]] = val[-1]
                    else:
                        parsed[key] = parsed[key][-1]
    return parsed


def parse_string_to_python(val):
    """
    Args:
        val: An input string

    Returns:
        A python object version of the input string
    """
    if val is None or len(val) == 0:
        return ""
    try:
        return literal_eval(val)
    except (ValueError, SyntaxError):
        try:
            return json.loads(val)
        except json.JSONDecodeError:
            return val


def parse_cli_to_dictionary(input_list):
    """
    Args:
        input_list: A list of input strings from a cli

    Returns:
        A dictionary constructed from the input list, with values converted to python objects where applicable
    """
    result = {}
    if input_list is None:
        return result
    key = ""
    val = ""
    idx = 0
    while idx < len(input_list):
        if input_list[idx].startswith("--"):
            if len(key) > 0:
                result[key] = parse_string_to_python(val)
            val = ""
            key = input_list[idx].strip('--')
        else:
            val += input_list[idx]
        idx += 1
    if len(key) > 0:
        result[key] = parse_string_to_python(val)
    return result


def convert_tf_dtype(datatype):
    """
    Gets the tensorflow datatype from string

    Args:
        datatype: String of datatype

    Returns:
        Tensor data type
    """
    datatype_map = {
        "string": tf.string,
        "int8": tf.int8,
        "uint8": tf.uint8,
        "int16": tf.int16,
        "uint16": tf.uint16,
        "int32": tf.int32,
        "uint32": tf.uint32,
        "int64": tf.int64,
        "uint64": tf.uint64,
        "float16": tf.float16,
        "float32": tf.float32,
        "float64": tf.float64
    }
    return datatype_map[datatype]


def strip_suffix(target, suffix):
    """
    Remove the given suffix from the target if it is present there

    Args:
        target: A string to be formatted
        suffix: A string to be removed from 'target'

    Returns:
        The formatted version of 'target'
    """
    if suffix is None or target is None:
        return target
    s_len = len(suffix)
    if target[-s_len:] == suffix:
        return target[:-s_len]
    return target


def prettify_metric_name(metric):
    """
    Add spaces to camel case words, then swap _ for space, and capitalize each word

    Args:
        metric: A string to be formatted

    Returns:
        The formatted version of 'metric'
    """
    return string.capwords(re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", metric).replace("_", " "))


def remove_blacklist_keys(dic, blacklist):
    """
    A function which removes the blacklisted elements from a dictionary

    Args:
        dic: The dictionary to inspect
        blacklist: keys to be removed from dic if they are present

    Returns:
        None

    Side Effects:
        Entries in dic may be removed
    """
    if blacklist is None:
        return
    for elem in blacklist:
        dic.pop(elem, None)


def is_number(s):
    """
    Args:
        s: A string
    Returns:
        True iff the string represents a number
    """
    try:
        float(s)
        return True
    except (ValueError, TypeError):
        return False


def decode_predictions(predictions, top=3, dictionary=None):
    """
    Args:
        predictions: A batched numpy array of class prediction scores (Batch X Predictions)
        top: How many of the highest predictions to capture
        dictionary: {"<class_idx>" -> "<class_name>"}
    Returns:
        A right-justified newline-separated array of the top classes and their associated probabilities.
        There is one entry in the results array per batch in the input
    """
    results = []
    for prediction in predictions:
        top_indices = prediction.argsort()[-top:][::-1]
        if dictionary is None:
            result = ["Class {:d}: {:.4f}".format(i, prediction[i]) for i in top_indices]
        else:
            result = ["{:s}: {:.4f}".format(dictionary[str(i)], prediction[i]) for i in top_indices]
        max_width = len(max(result, key=lambda s: len(s)))
        result = str.join("\n", [s.rjust(max_width) for s in result])
        results.append(result)
    return results


class Suppressor(object):
    """
    A class which can be used to silence output of function calls. example:
    with Suppressor():
        func(args)
    """
    def __enter__(self):
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        if exc_type is not None:
            raise

    def write(self, x):
        pass


class Timer(ContextDecorator):
    """
    A class that can be used to time things:
    with Timer():
        func(args)

    @Timer()
    def func(args)
    """
    def __init__(self, name="Task"):
        self.name = name

    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = time.perf_counter()
        self.interval = self.end - self.start
        tf.print("{} took {} seconds".format(self.name, self.interval))


class NonContext(object):
    """
    A class which is used for nothing
    """
    def __enter__(self):
        pass

    def __exit__(self, *exc):
        pass


def get_num_devices():
    local_device_protos = device_lib.list_local_devices()
    gpu_list = [x.name for x in local_device_protos if x.device_type == 'GPU']
    return max(1, len(gpu_list))


def flatten_list(input_list):
    for idx, ele in enumerate(input_list):
        input_list[idx] = to_list(ele)
    output_list = list(chain.from_iterable(input_list))
    return output_list


def to_list(data):
    if not isinstance(data, list):
        if isinstance(data, (tuple, set)):
            data = list(data)
        else:
            data = [data]
    return data



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import sys


def predict(args, unknown):
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    print("Prediction not yet supported, but coming soon...!")


def configure_predict_parser(subparsers):
    parser = subparsers.add_parser('predict', description='Evaluate a trained model on new inputs', allow_abbrev=False)
    parser.set_defaults(func=predict)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from collections import ChainMap, deque

import numpy as np
import tensorflow as tf

from fastestimator.estimator.trace import Logger, LRController, ModelSaver, MonitorLoss, Trace, TrainInfo
from fastestimator.network.lrschedule import LRSchedule
from fastestimator.util.cli_util import draw
from fastestimator.util.util import get_num_devices


class Estimator:
    def __init__(self,
                 pipeline,
                 network,
                 epochs,
                 steps_per_epoch=None,
                 validation_steps=None,
                 traces=None,
                 log_steps=100):
        self.pipeline = pipeline
        self.network = network
        self.epochs = epochs
        self.steps_per_epoch = steps_per_epoch
        self.validation_steps = validation_steps
        self.traces = traces
        assert log_steps is None or log_steps > 0, "log_steps must be positive or None"
        self.log_steps = log_steps
        self.inputs = None
        self.num_devices = get_num_devices()
        if self.num_devices > 1:
            self.distribute_strategy = tf.distribute.MirroredStrategy()
        else:
            self.distribute_strategy = None
        self.train_step = 0
        self.train_epoch = 0
        self.total_train_steps = 0
        self.do_eval = False

    def fit(self):
        """
        Function to perform training on the estimator
        """
        draw()
        self._prepare_pipeline()
        self._prepare_network()
        self._warmup()
        self._prepare_estimator()
        self._start()

    def _prepare_pipeline(self):
        self.pipeline.global_batch_multiplier = get_num_devices()
        self.pipeline.prepare(distribute_strategy=self.distribute_strategy)
        self.do_eval = "eval" in self.pipeline.mode_list

    def _prepare_network(self):
        self.network.num_devices = self.num_devices
        self.network.prepare(mode_list=self.pipeline.mode_list, distribute_strategy=self.distribute_strategy)

    def _prepare_estimator(self):
        if self.traces is None:
            self.traces = []
        elif not isinstance(self.traces, list):
            self.traces = [self.traces]
        self._add_traces()
        no_save_warning = True
        for trace in self.traces:
            assert isinstance(trace, Trace)
            trace.network = self.network
            if isinstance(trace, ModelSaver):
                no_save_warning = False
        if no_save_warning:
            print("FastEstimator-Warn: No ModelSaver Trace detected. Models will not be saved.")
        self._sort_traces()

    def _add_traces(self):
        if self.log_steps:
            self.traces.insert(0, TrainInfo(log_steps=self.log_steps))
            if not any(map(lambda x: isinstance(x, Logger), self.traces)):
                self.traces.append(Logger(log_steps=self.log_steps))
        self.traces.insert(0, MonitorLoss())

    def _sort_traces(self):
        # This is essentially a topological sort, but it doesn't seem worthwhile to convert the data into a graph
        # representation in order to get the slightly better asymptotic runtime complexity
        sorted_traces = []
        available_outputs = {
            "num_devices", "mode", "epoch", "train_step", "batch_idx", "batch_size", "batch", "elapsed_time"
        } | self.pipeline.all_output_keys | self.network.all_output_keys
        end_traces = deque()
        intermediate_traces = deque()
        intermediate_outputs = set()
        trace_deque = deque(self.traces)
        while len(trace_deque) > 0:
            trace = trace_deque.popleft()
            if not trace.inputs:
                sorted_traces.append(trace)
                available_outputs |= trace.outputs
            elif "*" in trace.inputs:
                if trace.outputs:
                    end_traces.appendleft(trace)
                else:
                    end_traces.append(trace)
            elif trace.inputs <= available_outputs:
                sorted_traces.append(trace)
                available_outputs |= trace.outputs
            else:
                intermediate_traces.append(trace)
                intermediate_outputs |= trace.outputs

        already_seen = set()
        while len(intermediate_traces) > 0:
            trace = intermediate_traces.popleft()
            already_seen.add(trace)
            if trace.inputs <= available_outputs:
                sorted_traces.append(trace)
                available_outputs |= trace.outputs
                already_seen.clear()
            elif trace.inputs <= (available_outputs | intermediate_outputs):
                intermediate_traces.append(trace)
            else:
                raise AssertionError("Trace {} has unsatisfiable inputs: {}".format(
                    type(trace).__name__, ", ".join(trace.inputs - (available_outputs | intermediate_outputs))))
            if 0 < len(already_seen) == len(intermediate_traces):
                raise AssertionError("Dependency cycle detected amongst traces: {}".format(", ".join(
                    [type(tr).__name__ for tr in already_seen])))
        sorted_traces.extend(list(end_traces))
        self.traces = sorted_traces

    def _warmup(self):
        mode_list = self.pipeline.mode_list
        self.total_train_steps = 0
        for mode in mode_list:
            epochs_pipeline = self.pipeline.dataset_schedule[mode].keys
            epochs_network = self.network.op_schedule[mode].keys
            signature_epochs = list(set(epochs_pipeline) | set(epochs_network))
            if mode == "train":
                elapse_epochs = np.diff(signature_epochs + [self.epochs])
                assert np.all(elapse_epochs > 0), "signature epoch is not sorted correctly"
            state = {"mode": mode}
            for idx, epoch in enumerate(signature_epochs):
                ds_iter = self.pipeline.dataset_schedule[mode].get_current_value(epoch)
                if mode == "train":
                    global_batch_size = self.pipeline.get_global_batch_size(epoch)
                    if self.steps_per_epoch:
                        max_steps = self.steps_per_epoch
                    else:
                        max_steps = min(self.pipeline.num_examples[mode]) // global_batch_size
                    self.total_train_steps += max_steps * elapse_epochs[idx]
                batch = next(ds_iter)
                state["batch_size"] = self.pipeline.get_global_batch_size(epoch)
                ops, model_list, epoch_losses = self.network.load_epoch(epoch, mode)
                if self.distribute_strategy:
                    self.distribute_strategy.experimental_run_v2(
                        self.network.run_step, args=(
                            batch,
                            ops,
                            model_list,
                            epoch_losses,
                            state,
                            True, ))
                else:
                    self.network.run_step(batch, ops, model_list, epoch_losses, state, warm_up=True)

    def _start(self):
        self.train_step = 0
        self._run_traces_on_begin({
            "train_step": self.train_step,
            "num_devices": self.num_devices,
            "log_steps": self.log_steps,
            "total_epochs": self.epochs,
            "total_train_steps": self.total_train_steps
        })
        for self.train_epoch in range(self.epochs):
            self._run_epoch("train")
            if self.do_eval:
                self._run_epoch("eval")
        self._run_traces_on_end({
            "train_step": self.train_step, "epoch": self.train_epoch, "num_devices": self.num_devices
        })

    def _run_epoch(self, mode):
        ds_iter = self.pipeline.dataset_schedule[mode].get_sequential_value(self.train_epoch)
        global_batch_size = self.pipeline.get_global_batch_size(self.train_epoch)
        if self.steps_per_epoch:
            max_steps = self.steps_per_epoch
        else:
            max_steps = min(self.pipeline.num_examples[mode]) // global_batch_size
        ops, model_list, epoch_losses = self.network.load_epoch(self.train_epoch, mode)
        self._run_traces_on_epoch_begin({"mode": mode, "epoch": self.train_epoch, "train_step": self.train_step})
        for batch_idx in range(max_steps):
            batch = next(ds_iter)
            self._run_traces_on_batch_begin({
                "mode": mode,
                "epoch": self.train_epoch,
                "train_step": self.train_step,
                "batch_idx": batch_idx,
                "batch_size": global_batch_size
            })
            if self.distribute_strategy:
                prediction, batch = self._forward_step_parallel(batch, ops,
                                                                model_list,
                                                                epoch_losses, {
                                                                    "mode": mode, "batch_size": global_batch_size
                                                                })
            else:
                prediction = self._forward_step(batch,
                                                ops,
                                                model_list,
                                                epoch_losses, {
                                                    "mode": mode, "batch_size": global_batch_size
                                                })
            batch = ChainMap(prediction, batch)
            self._run_traces_on_batch_end({
                "mode": mode,
                "epoch": self.train_epoch,
                "train_step": self.train_step,
                "batch_idx": batch_idx,
                "batch_size": global_batch_size,
                "batch": batch,
            })
            if mode == "train":
                self.train_step += 1
        self._run_traces_on_epoch_end({"mode": mode, "epoch": self.train_epoch, "train_step": self.train_step})

    def _run_traces_on_begin(self, state):
        trace_outputs = {}
        trace_state = ChainMap(trace_outputs, state)
        for trace in self.traces:
            trace.on_begin(trace_state)
        self._check_early_exit()

    def _run_traces_on_epoch_begin(self, state):
        trace_outputs = {}
        trace_state = ChainMap(trace_outputs, state)
        for trace in self.traces:
            if trace.mode is None or state['mode'] in trace.mode:
                trace.on_epoch_begin(trace_state)
        self._check_early_exit()

    def _run_traces_on_batch_begin(self, state):
        trace_outputs = {}
        trace_state = ChainMap(trace_outputs, state)
        for trace in self.traces:
            if trace.mode is None or state['mode'] in trace.mode:
                trace.on_batch_begin(trace_state)
        self._check_early_exit()

    def _run_traces_on_batch_end(self, state):
        trace_outputs = {}
        trace_state = ChainMap(trace_outputs, state)
        for trace in self.traces:
            if trace.mode is None or state['mode'] in trace.mode:
                trace.on_batch_end(trace_state)
        self._check_early_exit()

    def _run_traces_on_epoch_end(self, state):
        trace_outputs = {}
        trace_state = ChainMap(trace_outputs, state)
        for trace in self.traces:
            if trace.mode is None or state['mode'] in trace.mode:
                trace.on_epoch_end(trace_state)
        self._check_early_exit()

    def _run_traces_on_end(self, state):
        trace_outputs = {}
        trace_state = ChainMap(trace_outputs, state)
        for trace in self.traces:
            trace.on_end(trace_state)

    def _check_early_exit(self):
        if self.network.stop_training:
            self._run_traces_on_end({
                "train_step": self.train_step,
                "train_epoch": self.train_epoch,
                "num_devices": self.num_devices, })
            exit(0)

    @tf.function
    def _forward_step(self, batch, ops, model_list, epoch_losses, state):
        prediction = self.network.run_step(batch, ops, model_list, epoch_losses, state)
        return prediction

    @tf.function
    def _forward_step_parallel(self, batch, ops, model_list, epoch_losses, state):
        prediction = self.distribute_strategy.experimental_run_v2(self.network.run_step,
                                                                  args=(
                                                                      batch,
                                                                      ops,
                                                                      model_list,
                                                                      epoch_losses,
                                                                      state, ))
        prediction = self._per_replica_to_global(prediction)
        batch = self._per_replica_to_global(batch)
        return prediction, batch

    @staticmethod
    def _per_replica_to_global(data):
        new_data = {}
        for key, value in data.items():
            new_data[key] = tf.concat(value.values, axis=0)
        return new_data



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import os
import re
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from scipy.ndimage.filters import gaussian_filter1d

from fastestimator.util.util import prettify_metric_name, remove_blacklist_keys, strip_suffix
from fastestimator.util.loader import PathLoader


def parse_file(file_path):
    """
    A function which will parse log files into a dictionary of metrics

    Args:
        file_path: The path to a log file

    Returns:
        A dictionary of the form {"metric_name":[[<step>,<value>],...]}
    """
    metrics = defaultdict(list)
    with open(file_path) as file:
        for line in file:
            if not line.startswith("FastEstimator-"):
                continue
            parsed_line = re.findall(r"([^:^;\s]+):[\s]*([-]?[0-9]+[.]?[0-9]*);", line)
            step = parsed_line[0]
            assert step[0] == "step", \
                "Log file (%s) seems to be missing step information, or step is not listed first" % file
            for metric in parsed_line[1:]:
                metrics[metric[0]].append([int(step[1]), float(metric[1])])
    return metrics


def graph_metrics(files_metrics, smooth_factor, save, save_path, share_legend, pretty_names):
    """
    A function which will plot (and optionally save) file metrics for comparison viewing / analysis

    Args:
        files_metrics: A dictionary of the form {"file_name":{"metric_name":[[<step>,<value>],...]}}
        smooth_factor: A non-negative float representing the magnitude of gaussian smoothing to apply (zero for none)
        save: Whether to save (true) or display (false) the generated graph
        save_path: Where to save the image if save is true
        share_legend: Whether to have one legend across all graphs (true) or one legend per graph (false)
        pretty_names: Whether to modify the metric names in graph titles (true) or leave them alone (false)

    Returns:
        None

    Side Effects:
        Generates and either displays or saves a graph file
    """
    assert smooth_factor >= 0, "smooth factor must be non-negative (zero to disable)"
    assert len(files_metrics) > 0, "must have at least one file to plot"

    metric_keys = {keys for files in files_metrics.values() for keys in files.keys()}
    metric_list = list(metric_keys)
    metric_list.sort()  # Sort the metrics alphabetically for consistency
    num_metrics = len(metric_list)
    num_experiments = len(files_metrics)

    assert num_metrics > 0, "must have at least one metric to plot"

    # map the metrics into an n x n grid, then remove any extra rows. Final grid will be m x n with m <= n
    num_cols = math.ceil(math.sqrt(num_metrics))
    metric_grid_location = {key: (idx // num_cols, idx % num_cols) for (idx, key) in enumerate(metric_list)}
    num_rows = math.ceil(num_metrics / num_cols)

    sns.set_context('paper')
    fig, axs = plt.subplots(num_rows, num_cols, sharex='all', figsize=(4 * num_cols, 2.8 * num_rows))

    # If only one row, need to re-format the axs object for consistency. Likewise for columns
    if num_rows == 1:
        axs = [axs]
        if num_cols == 1:
            axs = [axs]

    for metric in metric_grid_location.keys():
        axis = axs[metric_grid_location[metric][0]][metric_grid_location[metric][1]]
        axis.set_title(metric if not pretty_names else prettify_metric_name(metric))
        axis.ticklabel_format(axis='y', style='sci', scilimits=(-2, 3))
        axis.grid(linestyle='--')
        axis.spines['top'].set_visible(False)
        axis.spines['right'].set_visible(False)
        axis.spines['bottom'].set_visible(False)
        axis.spines['left'].set_visible(False)
        axis.tick_params(bottom=False, left=False)

    for i in range(num_cols):
        axs[num_rows - 1][i].set_xlabel('Steps')

    # some of the columns in the last row might be unused, so disable them
    last_column_idx = num_cols - (num_rows * num_cols - num_metrics) - 1
    for i in range(last_column_idx + 1, num_cols):
        axs[num_rows - 1][i].axis('off')
        axs[num_rows - 2][i].set_xlabel('Steps')
        axs[num_rows - 2][i].xaxis.set_tick_params(which='both', labelbottom=True)

    colors = sns.hls_palette(n_colors=num_experiments,
                             s=0.95) if num_experiments > 10 else sns.color_palette("colorblind")

    handles = []
    labels = []
    for (color_idx, file) in enumerate(files_metrics):
        labels.append(file)
        for (idx, metric) in enumerate(files_metrics[file]):
            data = np.array(files_metrics[file][metric])
            y = data[:, 1] if smooth_factor == 0 else gaussian_filter1d(data[:, 1], sigma=smooth_factor)
            ln = axs[metric_grid_location[metric][0]][metric_grid_location[metric][1]].plot(
                data[:, 0], y, color=colors[color_idx], label=file, linewidth=1.5)
            if idx == 0:
                handles.append(ln[0])

    plt.tight_layout()

    if share_legend and num_rows > 1:
        if last_column_idx == num_cols - 1:
            fig.subplots_adjust(bottom=0.15)
            fig.legend(handles, labels, loc='lower center', ncol=num_cols + 1)
        else:
            axs[num_rows - 1][last_column_idx + 1].legend(handles, labels, loc='center', fontsize='large')
    else:
        for i in range(num_rows):
            for j in range(num_cols):
                if i == num_rows - 1 and j > last_column_idx:
                    break
                axs[i][j].legend(loc='best', fontsize='small')

    if not save:
        plt.show()
    else:
        save_path = os.path.dirname(save_path)
        if save_path is None or save_path == "":
            save_path = "."
        os.makedirs(save_path, exist_ok=True)
        save_file = os.path.join(save_path, 'parse_logs.png')
        print("Saving to %s" % save_file)
        plt.savefig(save_file, dpi=300)


def parse_files(file_paths,
                log_extension='.txt',
                smooth_factor=0,
                save=False,
                save_path=None,
                ignore_metrics=None,
                share_legend=True,
                pretty_names=False):
    """
    A function which will iterate through the given log file paths, parse them to extract metrics, remove any
        metrics which are blacklisted, and then pass the necessary information on the graphing function

    Args:
        file_paths: A list of paths to various log files
        log_extension: The extension of the log files
        smooth_factor: A non-negative float representing the magnitude of gaussian smoothing to apply (zero for none)
        save: Whether to save (true) or display (false) the generated graph
        save_path: Where to save the image if save is true. Defaults to dir_path if not provided
        ignore_metrics: Any metrics within the log files which will not be visualized
        share_legend: Whether to have one legend across all graphs (true) or one legend per graph (false)
        pretty_names: Whether to modify the metric names in graph titles (true) or leave them alone (false)

    Returns:
        None
    """
    if file_paths is None or len(file_paths) < 1:
        raise AssertionError("must provide at least one log file")

    files_metrics = {}
    for file_path in file_paths:
        file_metrics = parse_file(file_path)
        remove_blacklist_keys(file_metrics, ignore_metrics)
        files_metrics[strip_suffix(os.path.split(file_path)[1].strip(), log_extension)] = file_metrics
    graph_metrics(files_metrics,
                  smooth_factor,
                  save,
                  save_path if save_path is not None else file_paths[0],
                  share_legend,
                  pretty_names)


def parse_folder(dir_path,
                 log_extension='.txt',
                 recursive_search=False,
                 smooth_factor=1,
                 save=False,
                 save_path=None,
                 ignore_metrics=None,
                 share_legend=True,
                 pretty_names=False):
    """
    A function which will gather all log files within a given folder and pass them along for visualization

    Args:
        dir_path: The path to a directory containing log files
        log_extension: The extension of the log files
        recursive_search: Whether to recursively search sub-directories for log files
        smooth_factor: A non-negative float representing the magnitude of gaussian smoothing to apply(zero for none)
        save: Whether to save (true) or display (false) the generated graph
        save_path: Where to save the image if save is true. Defaults to dir_path if not provided
        ignore_metrics: Any metrics within the log files which will not be visualized
        share_legend: Whether to have one legend across all graphs (true) or one legend per graph (false)
        pretty_names: Whether to modify the metric names in graph titles (true) or leave them alone (false)

    Returns:
        None
    """
    loader = PathLoader(dir_path, input_extension=log_extension, recursive_search=recursive_search)
    file_paths = [x[0] for x in loader.path_pairs]

    parse_files(file_paths, log_extension, smooth_factor, save, save_path, ignore_metrics, share_legend, pretty_names)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math

import tensorflow as tf

import tensorflow_probability as tfp
from fastestimator.util.op import TensorOp


class Augmentation2D(TensorOp):
    """
   This class supports commonly used 2D random affine transformations for data augmentation.
   Either a scalar ``x`` or a tuple ``[x1, x2]`` can be specified for rotation, shearing, shifting, and zoom.

   Args:
       rotation_range: Scalar (x) that represents the range of random rotation (in degrees) from -x to x /
        Tuple ([x1, x2]) that represents  the range of random rotation between x1 and x2.
       width_shift_range: Float (x) that represents the range of random width shift (in pixels) from -x to x /
        Tuple ([x1, x2]) that represents  the range of random width shift between x1 and x2.
       height_shift_range: Float (x) that represents the range of random height shift (in pixels) from -x to x /
        Tuple ([x1, x2]) that represents  the range of random height shift between x1 and x2.
       shear_range: Scalar (x) that represents the range of random shear (in degrees) from -x to x /
        Tuple ([x1, x2]) that represents  the range of random shear between x1 and x2.
       zoom_range: Float (x) that represents the range of random zoom (in percentage) from -x to x /
        Tuple ([x1, x2]) that represents  the range of random zoom between x1 and x2.
       flip_left_right: Boolean representing whether to flip the image horizontally with a probability of 0.5.
       flip_up_down: Boolean representing whether to flip the image vertically with a probability of 0.5.
       mode: Augmentation on 'training' data or 'evaluation' data.
   """
    def __init__(self,
                 inputs=None,
                 outputs=None,
                 mode=None,
                 rotation_range=0.,
                 width_shift_range=0.,
                 height_shift_range=0.,
                 shear_range=0.,
                 zoom_range=1.,
                 flip_left_right=False,
                 flip_up_down=False):
        super().__init__(inputs, outputs, mode)
        self.rotation_range = rotation_range
        self.width_shift_range = width_shift_range
        self.height_shift_range = height_shift_range
        self.shear_range = shear_range
        self.zoom_range = zoom_range
        self.flip_left_right_boolean = flip_left_right
        self.flip_up_down_boolean = flip_up_down
        self.transform_matrix = tf.eye(3)
        self.width = None
        self.height = None
        self.do_flip_lr_tensor = tf.convert_to_tensor(0)
        self.do_flip_up_tensor = tf.convert_to_tensor(0)

    def _get_transform(self):
        return self.transform_matrix

    def rotate(self):
        """
        Creates affine transformation matrix for 2D rotation

        Returns:
            Transform affine tensor
        """
        rotation_range = [0., 0.]
        if type(self.rotation_range) is not tuple and type(self.rotation_range) is not list:
            rotation_range[0] = -1 * self.rotation_range
            rotation_range[1] = self.rotation_range
        else:
            rotation_range = self.rotation_range
        self.rotation_range = rotation_range
        theta = tf.random.uniform([],
                                  maxval=math.pi / 180 * self.rotation_range[1],
                                  minval=math.pi / 180 * self.rotation_range[0])
        base_matrix = tf.constant([[1, 0, 0], [0, 1, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)
        rotation_matrix_1 = tf.cos(theta) * base_matrix
        base_matrix = tf.constant([[0, -1, 0], [1, 0, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)
        rotation_matrix_2 = base_matrix * tf.sin(theta)
        transform_matrix = rotation_matrix_1 + rotation_matrix_2 + tf.constant(
            [[0, 0, 0], [0, 0, 0], [0, 0, 1]], shape=[3, 3], dtype=tf.float32)
        transform_matrix = self.transform_matrix_offset_center(transform_matrix)
        return transform_matrix

    def shift(self):
        """
        Creates affine transformation matrix for 2D shift

        Returns:
            Transform affine tensor
        """
        width_shift_range = [0., 0.]
        if type(self.rotation_range) is not tuple and type(self.width_shift_range) is not list:
            width_shift_range[0] = -1 * self.width_shift_range
            width_shift_range[1] = self.width_shift_range
        else:
            width_shift_range = self.width_shift_range
        self.width_shift_range = width_shift_range
        height_shift_range = [0., 0.]
        if type(self.height_shift_range) is not tuple and type(self.height_shift_range) is not list:
            height_shift_range[0] = -1 * self.height_shift_range
            height_shift_range[1] = self.height_shift_range
        else:
            height_shift_range = self.height_shift_range
        self.height_shift_range = height_shift_range
        ty = tf.random.uniform([], maxval=self.width_shift_range[1], minval=self.width_shift_range[0])
        ty *= self.width
        base_ty = ty * tf.constant([[0, 0, 1], [0, 0, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)
        tx = tf.random.uniform([], maxval=self.height_shift_range[1], minval=self.height_shift_range[0])
        tx *= self.height
        base_tx = tx * tf.constant([[0, 0, 0], [0, 0, 1], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)
        transform_matrix = tf.constant([[1, 0, 0], [0, 1, 0], [0, 0, 1]], shape=[3, 3],
                                       dtype=tf.float32) + base_tx + base_ty
        transform_matrix = self.transform_matrix_offset_center(transform_matrix)
        return transform_matrix

    def shear(self):
        """
        Creates affine transformation matrix for 2D shear

        Returns:
            Transform affine tensor
        """
        shear_range = [0., 0.]
        if type(self.shear_range) is not tuple and type(self.shear_range) is not list:
            shear_range[0] = -1 * self.shear_range
            shear_range[1] = self.shear_range
        else:
            shear_range = self.shear_range
        self.shear_range = shear_range
        sx = tf.random.uniform([],
                               maxval=math.pi / 180 * self.shear_range[1],
                               minval=math.pi / 180 * self.shear_range[0])
        sy = tf.random.uniform([],
                               maxval=math.pi / 180 * self.shear_range[1],
                               minval=math.pi / 180 * self.shear_range[0])
        base_shear1 = -tf.sin(sx) * tf.constant([[0, 1, 0], [0, 0, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)
        base_shear2 = tf.cos(sy) * tf.constant([[0, 0, 0], [0, 1, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)
        transform_matrix = tf.constant([[1, 0, 0], [0, 0, 0], [0, 0, 1]], shape=[3, 3], dtype=tf.float32) + \
                           base_shear1 + base_shear2
        transform_matrix = self.transform_matrix_offset_center(transform_matrix)
        return transform_matrix

    def zoom(self):
        """
        Creates affine transformation matrix for 2D zoom / scale

        Returns:
            Transform affine tensor
        """
        zoom_range = [0., 0.]
        if type(self.zoom_range) is not tuple and type(self.zoom_range) is not list:
            zoom_range[0] = 1 - self.zoom_range
            zoom_range[1] = 1 + self.zoom_range
        else:
            zoom_range = self.zoom_range
        self.zoom_range = zoom_range
        zx = tf.random.uniform([], maxval=self.zoom_range[1], minval=self.zoom_range[0])
        zy = tf.random.uniform([], maxval=self.zoom_range[1], minval=self.zoom_range[0])
        base_zx = tf.constant([[1, 0, 0], [0, 0, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32) / zx
        base_zy = tf.constant([[0, 0, 0], [0, 1, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32) / zy
        transform_matrix = tf.constant([[0, 0, 0], [0, 0, 0], [0, 0, 1]], shape=[3, 3],
                                       dtype=tf.float32) + base_zx + base_zy
        transform_matrix = self.transform_matrix_offset_center(transform_matrix)
        return transform_matrix

    def flip(self):
        """
        Decides whether or not to flip

        Returns:
            A boolean that represents whether or not to flip
        """
        return tf.cond(
            tf.random.uniform([], minval=0, maxval=1) > 0.5,
            lambda: tf.convert_to_tensor(1),
            lambda: tf.convert_to_tensor(0))

    def transform_matrix_offset_center(self, matrix):
        """
        Offsets the tensor to the center of the image

        Args:
            matrix: Affine tensor

        Returns:
            An affine tensor offset to the center of the image
        """
        o_x = self.height / tf.constant([2], dtype=tf.float32) + tf.constant([0.5], dtype=tf.float32)
        o_y = self.width / tf.constant([2], dtype=tf.float32) + tf.constant([0.5], dtype=tf.float32)
        eye = tf.constant([[1, 0, 0], [0, 1, 0], [0, 0, 1]], shape=[3, 3], dtype=tf.float32)

        offset_matrix = eye + \
                        tf.multiply(o_x,
                                    tf.constant([[0, 0, 1], [0, 0, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)) + \
                        tf.multiply(o_y, tf.constant([[0, 0, 0], [0, 0, 1], [0, 0, 0]], shape=[3, 3], dtype=tf.float32))

        reset_matrix = eye + \
                       tf.multiply(o_x,
                                   tf.constant([[0, 0, -1], [0, 0, 0], [0, 0, 0]], shape=[3, 3], dtype=tf.float32)) + \
                       tf.multiply(tf.constant([[0, 0, 0], [0, 0, -1], [0, 0, 0]], shape=[3, 3], dtype=tf.float32), o_y)

        transform_matrix = tf.tensordot(tf.tensordot(offset_matrix, matrix, axes=1), reset_matrix, axes=1)
        return transform_matrix

    def setup(self):
        """
        This method set the appropriate variables necessary for the random 2D augmentation. It also computes the
        transformation matrix.

        Returns:
            None
        """
        # \NOTE(JP): tracing behavior from dataset.map causes issue when any tensor id defined as tf.constant
        assert self.width is not None, "width is not set yet"
        assert self.height is not None, "height is not set yet"
        transform_matrix = tf.convert_to_tensor([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=tf.float32)
        do_rotate = False
        do_shift = False
        do_zoom = False
        do_shear = False
        self.do_flip_lr_tensor = tf.convert_to_tensor(0)
        self.do_flip_ud_tensor = tf.convert_to_tensor(0)

        if type(self.rotation_range) is not tuple and type(self.rotation_range) is not list:
            if self.rotation_range > 0.:
                do_rotate = True
        else:
            if self.rotation_range[0] > 0. or self.rotation_range[1] > 0.:
                do_rotate = True

        if type(self.width_shift_range) is not tuple and type(self.width_shift_range) is not list:
            if self.width_shift_range > 0.:
                do_shift = True
        else:
            if self.width_shift_range[0] > 0. or self.width_shift_range[1] > 0.:
                do_shift = True

        if type(self.height_shift_range) is not tuple and type(self.height_shift_range) is not list:
            if self.height_shift_range > 0.:
                do_shift = True
        else:
            if self.height_shift_range[0] > 0. or self.height_shift_range[1] > 0.:
                do_shift = True

        if type(self.zoom_range) is not tuple and type(self.zoom_range) is not list:
            if self.zoom_range != 1.:
                do_zoom = True
        else:
            if self.zoom_range[0] != 1. or self.zoom_range[1] != 0.:
                do_zoom = True

        if type(self.shear_range) is not tuple and type(self.shear_range) is not list:
            if self.shear_range > 0.:
                do_shear = True
        else:
            if self.shear_range[0] > 0. or self.shear_range[1] > 0.:
                do_shear = True

        if do_rotate:
            if transform_matrix is None:
                transform_matrix = self.rotate()
            else:
                transform_matrix = tf.tensordot(transform_matrix, self.rotate(), axes=1)

        if do_shift:
            if transform_matrix is None:
                transform_matrix = self.shift()
            else:
                transform_matrix = tf.tensordot(transform_matrix, self.shift(), axes=1)

        if do_zoom:
            if transform_matrix is None:
                transform_matrix = self.zoom()
            else:
                transform_matrix = tf.tensordot(transform_matrix, self.zoom(), axes=1)
        if do_shear:
            if transform_matrix is None:
                transform_matrix = self.shear()
            else:
                transform_matrix = tf.tensordot(transform_matrix, self.shear(), axes=1)

        self.transform_matrix = transform_matrix

        if self.flip_left_right_boolean:
            self.do_flip_lr_tensor = self.flip()
        if self.flip_up_down_boolean:
            self.do_flip_ud_tensor = self.flip()

    def forward(self, data, state):
        """
        Transforms the data with the augmentation transformation

        Args:
            data: Data to be transformed
            state: Information about the current execution context

        Returns:
            Transformed (augmented) data

        """
        # ensure the data is list in order to prevent syntax error at 322
        if not isinstance(data, list):
            if isinstance(data, tuple):
                data = list(data)
            else:
                data = [data]
        self.width = tf.cast(data[0].shape[-3], tf.float32)
        self.height = tf.cast(data[0].shape[-2], tf.float32)

        self.setup()
        for idx, single_data in enumerate(data):
            augment_data = self._transform(single_data)
            augment_data = tf.cond(tf.equal(self.do_flip_lr_tensor, 1),
                                   lambda: tf.image.flip_left_right(augment_data),
                                   lambda: augment_data)
            augment_data = tf.cond(tf.equal(self.do_flip_ud_tensor, 1),
                                   lambda: tf.image.flip_up_down(augment_data),
                                   lambda: augment_data)
            data[idx] = augment_data
        return data

    def _transform(self, data):
        dtype = data.dtype

        assert (self.width is not None)
        assert (self.height is not None)

        x_shape = self.width
        y_shape = self.height
        z_shape = tf.cast(data.shape[-1], tf.float32)

        x_range = tf.range(x_shape)
        y_range = tf.range(y_shape)
        z_range = tf.range(z_shape)
        x_, y_, z_ = tf.meshgrid(x_range, y_range, z_range, indexing='ij')

        x_ = tf.reshape(x_, [-1])
        y_ = tf.reshape(y_, [-1])
        z_ = tf.reshape(z_, [-1])

        coords = tf.stack([x_, y_, tf.ones_like(x_)])
        M = tf.linalg.inv(self.transform_matrix)
        coords = tf.matmul(M, coords)

        x_ = coords[0]
        y_ = coords[1]

        x_ = tf.cast(tf.clip_by_value(tf.round(x_), 0, x_shape - 1), tf.int32)
        y_ = tf.cast(tf.clip_by_value(tf.round(y_), 0, y_shape - 1), tf.int32)
        z_ = tf.cast(z_, tf.int32)
        final_coords = tf.stack([x_, y_, z_], axis=-1)
        gather_and_reshape = lambda ex: tf.reshape(tf.gather_nd(ex, final_coords), ex.shape)

        if len(data.shape) > 3:
            result = tf.map_fn(fn=gather_and_reshape, elems=data, dtype=dtype)
        else:
            result = gather_and_reshape(data)
        return result

class MixUpBatch(TensorOp):
    """
    This class should be used in conjunction with MixUpLoss to perform mix-up training, which helps to reduce
    over-fitting, stabilize GAN training, and against adversarial attacks (https://arxiv.org/abs/1710.09412)
    """
    def __init__(self, inputs=None, outputs=None, mode=None, alpha=1.0):
        """
        Args:
            inputs: key of the input to be mixed up
            outputs: key to store the mixed-up input
            mode: what mode to execute in. Probably 'train'
            alpha: the alpha value defining the beta distribution to be drawn from during training
        """
        assert alpha > 0, "Mixup alpha value must be greater than zero"
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.alpha = tf.constant(alpha)
        self.beta = tfp.distributions.Beta(alpha, alpha)

    def forward(self, data, state):
        iterdata = data if isinstance(data, list) else list(data) if isinstance(data, tuple) else [data]
        lam = self.beta.sample()
        # Could do random mix-up using tf.gather() on a shuffled index list, but batches are already randomly ordered,
        # so just need to roll by 1 to get a random combination of inputs. This also allows MixUpLoss to easily compute
        # the corresponding Y values
        mix = [lam * dat + (1.0 - lam) * tf.roll(dat, shift=1, axis=0) for dat in iterdata]
        return mix + [lam]


class AdversarialSample(TensorOp):
    def __init__(self, inputs, outputs=None, mode=None, epsilon=0.1):
        assert len(inputs) == 2, \
            "AdversarialSample requires 2 inputs: a loss value and the input data which lead to the loss"
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.epsilon = epsilon

    def forward(self, data, state):
        clean_loss, clean_data = data
        tape = state['tape']
        with tape.stop_recording():
            grad_clean = tape.gradient(clean_loss, clean_data)
            adverse_data = tf.clip_by_value(clean_data + self.epsilon * tf.sign(grad_clean),
                                            tf.reduce_min(clean_data),
                                            tf.reduce_max(clean_data))
        return adverse_data


class Average(TensorOp):
    def forward(self, data, state):
        iterdata = data if isinstance(data, list) else list(data) if isinstance(data, tuple) else [data]
        num_entries = len(iterdata)
        result = 0.0
        for elem in iterdata:
            result += elem
        return result / num_entries



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
import os

# noinspection PyPackageRequirements
import tensorflow as tf

from fastestimator.util.util import load_image


class PathLoader(object):
    def __init__(self, root_path, batch=10, input_extension=None, recursive_search=True):
        """
        Args:
            root_path: The path the the root directory containing files to be read
            batch: The batch size to use when loading paths. Must be positive
            input_extension: A file extension to limit what sorts of paths are returned
            recursive_search: Whether to search within subdirectories for files
        """
        if not os.path.isdir(root_path):
            raise AssertionError("Provided path is not a directory")
        self.root_path = root_path
        if batch < 1:
            raise AssertionError("Batch size must be positive")
        self.batch = batch
        self.input_extension = input_extension
        self.recursive_search = recursive_search
        self.current_idx = 0
        self.path_pairs = self.get_file_paths()

    def __iter__(self):
        return self

    def get_file_paths(self):
        path_pairs = []
        for root, dirs, files in os.walk(self.root_path):
            for file_name in files:
                if file_name.startswith(".") or (self.input_extension is not None
                                                 and not file_name.endswith(self.input_extension)):
                    continue
                path_pairs.append((os.path.join(root, file_name), os.path.basename(root)))
            if not self.recursive_search:
                break
        return path_pairs

    def __next__(self):
        if self.current_idx >= len(self.path_pairs):
            raise StopIteration
        else:
            result = self.path_pairs[self.current_idx:self.current_idx + self.batch]
            self.current_idx += self.batch
            return result

    def __len__(self):
        return math.ceil(len(self.path_pairs) / self.batch)


class ImageLoader(PathLoader):
    def __init__(self, root_path, model, batch=10, input_extension=None, strip_alpha=False):
        super().__init__(root_path, batch, input_extension)
        self.strip_alpha = strip_alpha
        self.input_type = model.input.dtype
        self.input_shape = model.input.shape
        if not (3 <= len(self.input_shape) <= 4):
            raise AssertionError("Model must have 3 or 4 dimensions: (batch, x, y, [channels])")
        if self.input_shape[0] is not None:
            raise AssertionError("Model must take batch on axis zero")
        self.n_channels = 0 if len(self.input_shape) == 3 else self.input_shape[3]

    def __next__(self):
        paths = super().__next__()
        inputs = [
            load_image(paths[i][0], strip_alpha=self.strip_alpha, channels=self.n_channels) for i in range(len(paths))
        ]
        batch_inputs = tf.stack([
            tf.image.resize_with_crop_or_pad(tf.convert_to_tensor(im, dtype=self.input_type), self.input_shape[1],
                                             self.input_shape[2]) for im in inputs
        ], axis=0)

        batch_classes = [paths[i][1] for i in range(len(paths))]
        return batch_inputs, batch_classes



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
Download horse2zebra dataset from https://people.eecs.berkeley.edu/~taesung_park/CycleGAN/datasets/horse2zebra.zip
"""
import os
import zipfile
import tempfile
from glob import glob

import pandas as pd
import wget


def _create_csv(img, key_name, parent_path, csv_name):
    df = pd.DataFrame()
    df[key_name] = [img_filename.replace(parent_path, '.') for img_filename in img]
    df.to_csv(csv_name, index=False)


def load_data(path=None):
    if path is None:
        path = os.path.join(tempfile.gettempdir(), ".fe", "HORSE2ZEBRA")
    if not (os.path.exists(path)):
        os.makedirs(path)
    data_compressed_path = os.path.join(path, 'horse2zebra.zip')
    data_folder_path = os.path.join(path, 'horse2zebra')
    trainA_csv = os.path.join(path, 'trainA.csv')
    trainB_csv = os.path.join(path, 'trainB.csv')
    testA_csv = os.path.join(path, 'testA.csv')
    testB_csv = os.path.join(path, 'testB.csv')
    if not os.path.exists(data_compressed_path):
        print("Downloading data to {}:".format(path))
        wget.download('https://people.eecs.berkeley.edu/~taesung_park/CycleGAN/datasets/horse2zebra.zip', path)
    if not os.path.exists(data_folder_path):
        print('\nExtracting files...')
        with zipfile.ZipFile(data_compressed_path, 'r') as zip_file:
            zip_file.extractall(path)
    if not os.path.exists(trainA_csv):
        trainA_img = glob(os.path.join(path, 'horse2zebra', 'trainA', '*.jpg'))
        _create_csv(trainA_img, 'imgA', path, trainA_csv)
    if not os.path.exists(trainB_csv):
        trainB_img = glob(os.path.join(path, 'horse2zebra', 'trainB', '*.jpg'))
        _create_csv(trainB_img, 'imgB', path, trainB_csv)
    if not os.path.exists(testA_csv):
        testA_img = glob(os.path.join(path, 'horse2zebra', 'testA', '*.jpg'))
        _create_csv(testA_img, 'imgA', path, testA_csv)
    if not os.path.exists(testB_csv):
        testB_img = glob(os.path.join(path, 'horse2zebra', 'testB', '*.jpg'))
        _create_csv(testB_img, 'imgB', path, testB_csv)
    return trainA_csv, trainB_csv, testA_csv, testB_csv, path



import os
import re

from setuptools import find_packages, setup


def get_version():
    path = os.path.dirname(__file__)
    version_re = re.compile(r'''__version__ = ['"](.+)['"]''')
    with open(os.path.join(path, 'fastestimator', '__init__.py')) as f:
        init = f.read()
    return version_re.search(init).group(1)


setup(
    name="fastestimator",
    version=get_version(),
    description="Deep learning Application framework",
    packages=find_packages(),
    package_dir={'': '.'},
    long_description="FastEstimator is a high-level deep learning API. With the help of FastEstimator, you can easily \
                      build a high-performance deep learning model and run it anywhere.",
    author="FastEstimator Dev",
    url='https://github.com/fastestimator/fastestimator',
    license="Apache License 2.0",
    keywords="fastestimator tensorflow",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3", ],

    # Declare minimal set for installation
    install_requires=[
        'numpy',
        'pyfiglet',
        'pandas',
        'pillow',
        'sklearn',
        'wget',
        'matplotlib',
        'seaborn>= 0.9.0',
        'scipy',
        'pytest',
        'pytest-cov',
        'tensorflow-probability',
        'umap-learn',
        'tqdm'
    ],
    # Declare extra set for installation
    extras_require={},
    scripts=['bin/fastestimator'])



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""" Mendeley dataset API.
Kermany, Daniel; Zhang, Kang; Goldbaum, Michael (2018), "Labeled Optical Coherence Tomography (OCT) and Chest X-Ray Images for Classification", Mendeley Data, v2
http://dx.doi.org/10.17632/rscbjbr9sj.2

CC BY 4.0 licence:
https://creativecommons.org/licenses/by/4.0/
"""

import os
import tempfile
import zipfile
import pandas as pd
import wget
import requests, io

def img_data_constructor(data_folder, mode, csv_path):
    normal_cases_dir = os.path.join(data_folder, 'NORMAL')
    pneumonia_cases_dir = os.path.join(data_folder, 'PNEUMONIA')
    data = []
    for dirpath, _, filenames in os.walk(normal_cases_dir):
        for f in filenames:
            # Select only jpeg files:
            if f[-4:] == 'jpeg':
                img = os.path.abspath(os.path.join(dirpath, f))
                data.append((img, 0))
    for dirpath, _, filenames in os.walk(pneumonia_cases_dir):
        for f in filenames:
            if f[-4:] == 'jpeg':
                img = os.path.abspath(os.path.join(dirpath, f))
                data.append((img, 1))
    num_example = len(data)
    print("found %d number of examples for %s" % (num_example, mode))
    data = pd.DataFrame(data, columns=['x', 'y'], index=None)
    data = data.sample(frac=1.).reset_index(drop=True)
    data.to_csv(csv_path, index=False)
    return data


def load_data(path=None):
    if path is None:
        path = os.path.join(tempfile.gettempdir(), ".fe", "Mendeley")
    if not os.path.exists(path):
        os.makedirs(path)
    train_csv = os.path.join(path, "train_data.csv")
    test_csv = os.path.join(path, "test_data.csv")
    data_compressed_path = os.path.join(path, "ChestXRay2017.zip")
    data_folder_path = os.path.join(path, "chest_xray")
    if not os.path.exists(data_compressed_path):
        print("Downloading data to %s" % path)
        r=requests.get('https://data.mendeley.com/datasets/rscbjbr9sj/2/files/41d542e7-7f91-47f6-9ff2-dd8e5a5a7861/ChestXRay2017.zip?dl=1')
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(path)
    if not os.path.exists(data_folder_path):
        print("extracting data...")
        with zipfile.ZipFile(data_compressed_path, 'r') as zip_file:
            zip_file.extractall(path=path)
    if not os.path.exists(train_csv):
        train_folder = os.path.join(data_folder_path, "train")
        img_data_constructor(train_folder, "train", train_csv)
    if not os.path.exists(test_csv):
        test_folder = os.path.join(data_folder_path, "test")
        img_data_constructor(test_folder, "test", test_csv)
    return train_csv, test_csv, path


# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
from tensorflow.python.keras import Model, layers
from tensorflow.python.keras.initializers import RandomNormal


# Code borrowed from https://github.com/tensorflow/examples/blob/master/tensorflow_examples/models/pix2pix/pix2pix.py
class InstanceNormalization(layers.Layer):
    """Instance Normalization Layer (https://arxiv.org/abs/1607.08022)."""
    def __init__(self, epsilon=1e-5):
        super().__init__()
        self.epsilon = epsilon

    def get_config(self):
        return {'epsilon': self.epsilon}

    def build(self, input_shape):
        self.scale = self.add_weight(name='scale',
                                     shape=input_shape[-1:],
                                     initializer=tf.random_normal_initializer(0., 0.02),
                                     trainable=True)

        self.offset = self.add_weight(name='offset', shape=input_shape[-1:], initializer='zeros', trainable=True)

    def call(self, x):
        mean, variance = tf.nn.moments(x, axes=[1, 2], keepdims=True)
        inv = tf.math.rsqrt(variance + self.epsilon)
        normalized = (x - mean) * inv
        return self.scale * normalized + self.offset


# Code borrowd from https://stackoverflow.com/questions/50677544/reflection-padding-conv2d
class ReflectionPadding2D(layers.Layer):
    def __init__(self, padding=(1, 1)):
        self.padding = tuple(padding)
        self.input_spec = [layers.InputSpec(ndim=4)]
        super().__init__()

    def get_config(self):
        return {'padding': self.padding}

    def compute_output_shape(self, s):
        """ If you are using "channels_last" configuration"""
        return s[0], s[1] + 2 * self.padding[0], s[2] + 2 * self.padding[1], s[3]

    def call(self, x, mask=None):
        w_pad, h_pad = self.padding
        return tf.pad(x, [[0, 0], [h_pad, h_pad], [w_pad, w_pad], [0, 0]], 'REFLECT')


def _resblock(x0, num_filter=256, kernel_size=3):
    x = ReflectionPadding2D()(x0)
    x = layers.Conv2D(filters=num_filter, kernel_size=kernel_size, kernel_initializer=RandomNormal(mean=0,
                                                                                                   stddev=0.02))(x)
    x = InstanceNormalization()(x)
    x = layers.ReLU()(x)

    x = ReflectionPadding2D()(x)
    x = layers.Conv2D(filters=num_filter, kernel_size=kernel_size, kernel_initializer=RandomNormal(mean=0,
                                                                                                   stddev=0.02))(x)
    x = InstanceNormalization()(x)
    x = layers.Add()([x, x0])
    return x


def build_discriminator(input_shape=(256, 256, 3)):
    x0 = layers.Input(input_shape)
    x = layers.Conv2D(filters=64,
                      kernel_size=4,
                      strides=2,
                      padding='same',
                      kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x0)
    x = layers.LeakyReLU(0.2)(x)

    x = layers.Conv2D(filters=128,
                      kernel_size=4,
                      strides=2,
                      padding='same',
                      kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)

    x = InstanceNormalization()(x)
    x = layers.LeakyReLU(0.2)(x)

    x = layers.Conv2D(filters=256,
                      kernel_size=4,
                      strides=2,
                      padding='same',
                      kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)

    x = InstanceNormalization()(x)
    x = layers.LeakyReLU(0.2)(x)

    x = ReflectionPadding2D()(x)
    x = layers.Conv2D(filters=512, kernel_size=4, strides=1, kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)

    x = InstanceNormalization()(x)
    x = layers.LeakyReLU(0.2)(x)

    x = ReflectionPadding2D()(x)
    x = layers.Conv2D(filters=1, kernel_size=4, strides=1, kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)

    return Model(inputs=x0, outputs=x)


def build_generator(input_shape=(256, 256, 3), num_blocks=9):
    x0 = layers.Input(input_shape)

    x = ReflectionPadding2D(padding=(3, 3))(x0)
    x = layers.Conv2D(filters=64, kernel_size=7, strides=1, kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)

    x = InstanceNormalization()(x)
    x = layers.ReLU()(x)

    # downsample
    x = layers.Conv2D(filters=128,
                      kernel_size=3,
                      strides=2,
                      padding='same',
                      kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)
    x = InstanceNormalization()(x)
    x = layers.ReLU()(x)

    x = layers.Conv2D(filters=256,
                      kernel_size=3,
                      strides=2,
                      padding='same',
                      kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)
    x = InstanceNormalization()(x)
    x = layers.ReLU()(x)

    # residual
    for _ in range(num_blocks):
        x = _resblock(x)

    # upsample
    x = layers.Conv2DTranspose(filters=128,
                               kernel_size=3,
                               strides=2,
                               padding='same',
                               kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)
    x = InstanceNormalization()(x)
    x = layers.ReLU()(x)

    x = layers.Conv2DTranspose(filters=64,
                               kernel_size=3,
                               strides=2,
                               padding='same',
                               kernel_initializer=RandomNormal(mean=0, stddev=0.02))(x)
    x = InstanceNormalization()(x)
    x = layers.ReLU()(x)

    # final
    x = ReflectionPadding2D(padding=(3, 3))(x)
    x = layers.Conv2D(filters=3, kernel_size=7, activation='tanh', kernel_initializer=RandomNormal(mean=0,
                                                                                                   stddev=0.02))(x)

    return Model(inputs=x0, outputs=x)
